"""Core classes for representing and simulating Gaussian beams and optical setups.

Setups are represented as :class:`OpticalSetup` instances which propagate an initial :class:`Beam`
through a sequence of :class:`Lens` elements. This yields a piecewise defined beam radius made
from Gaussian beam segments between the elements. The individual beams are represented using
a complex beam parameter combined with an axial offset and wavelength. The beams are propagated
using the ray transfer matrix method for Gaussian beams, see `here <https://en.wikipedia.org/wiki/Ray_transfer_matrix_analysis#Gaussian_beams>`_.
"""

import warnings
from collections.abc import Sequence
from dataclasses import InitVar, dataclass
from functools import cached_property
from itertools import pairwise
from typing import ClassVar

import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import curve_fit

from .plot import OpticalSetupPlot, fig_to_png, plot_setup
from .serialize import YamlSerializableMixin


# TODO should beam include wavelength or should it be part of the larger setup?
# TODO refractive index?
@dataclass(frozen=True)
class Beam(YamlSerializableMixin):
    """Paraxial Gaussian beam representation.

    Implements :meth:`_repr_png_` to show a plot of the beam radius in IPython environments.
    """

    beam_parameter: complex
    """Complex beam parameter :math:`q = z - z_0 + i z_R` defined at position :math:`z`
    with focus at :math:`z_0` and Rayleigh range :math:`z_R`."""
    z_offset: float  #: Axial position at which the ray is defined
    wavelength: float  #: Wavelength of the beam
    gauss_cov: np.ndarray | None = None  #: 2x2 covariance matrix for focus position and waist
    # range: tuple[float, float] # TODO is this necessary

    @cached_property
    def waist(self) -> float:
        """Waist radius"""
        return np.sqrt(self.rayleigh_range * self.wavelength / np.pi)

    @cached_property
    def rayleigh_range(self) -> float:
        """Rayleigh range"""
        return abs(self.beam_parameter.imag)

    @cached_property
    def focus(self) -> float:
        """Axial position of the beam focus i.e. waist position"""
        return self.z_offset - self.beam_parameter.real

    def radius(self, z: float | np.ndarray) -> float | np.ndarray:
        """Compute the beam radius at axial position(s).

        Args:
            z: Axial position(s) where the beam radius is evaluated.

        Returns:
            Beam radius at the specified axial position(s).
        """
        return self.waist * np.sqrt(1 + ((z - self.focus) / self.rayleigh_range) ** 2)

    def radius_dev(self, z: float | np.ndarray) -> float | np.ndarray:
        """Compute the beam radius (standard) deviation at axial position(s).

        Args:
            z: Axial position(s) where the beam radius deviation is evaluated.

        Returns:
            Beam radius deviation at the specified axial position(s).

        Raises:
            ValueError: If the covariance matrix is not defined for this beam.
        """
        # derivation:
        """
        >>> import sympy as sp
        >>> z, focus = sp.symbols('z focus', real=True)
        >>> waist, wavelength = sp.symbols('self.waist self.wavelength', real=True, positive=True)
        >>> rayleigh_range = sp.pi * waist**2 / wavelength
        >>> radius = waist * sp.sqrt(1 + ((z - focus) / rayleigh_range) ** 2)
        >>> res = radius.diff(sp.Matrix([focus, waist])).simplify()
        >>> print(res[0])
        self.wavelength**2*(focus - z)/(pi*self.waist*sqrt(pi**2*self.waist**4 + self.wavelength**2*(focus - z)**2))
        >>> print(res[1])
        (pi**2*self.waist**4 - self.wavelength**2*(focus - z)**2)/(pi*self.waist**2*sqrt(pi**2*self.waist**4 + self.wavelength**2*(focus - z)**2))
        """
        if self.gauss_cov is None:
            raise ValueError("Covariance matrix is not defined for this beam.")
        wl = self.wavelength
        common_denom = np.pi * self.waist * np.sqrt(np.pi**2 * self.waist**4 + wl**2 * (self.focus - z) ** 2)
        jac = np.array(
            [
                [wl**2 * (self.focus - z) / common_denom],
                [(np.pi**2 * self.waist**4 - wl**2 * (self.focus - z) ** 2) / common_denom / self.waist],
            ]
        )
        if np.isscalar(z):
            return float(np.sqrt(jac.T @ self.gauss_cov @ jac))
        else:
            return np.sqrt(np.einsum("ij,jk,ik->i", jac, self.gauss_cov, jac))

    @classmethod
    def from_gauss(cls, focus: float, waist: float, wavelength: float, cov: np.ndarray | None = None) -> "Beam":
        """Create a Beam instance from focus position and waist.

        Args:
            focus: The axial position of the beam focus.
            waist: The beam waist radius.
            wavelength: The wavelength of the beam.
            cov: Optional 2x2 covariance matrix for the focus and waist.
        Returns:
            Beam instance.
        """

        rayleigh_range = np.pi * (waist**2) / wavelength
        return cls(beam_parameter=1j * rayleigh_range, z_offset=focus, wavelength=wavelength, gauss_cov=cov)

    @classmethod
    def fit(
        cls, positions: np.ndarray, radii: np.ndarray, wavelength: float, p0: tuple[float, float] | None = None
    ) -> "Beam":
        """Fit a Gaussian beam radius to measured data.

        This uses :func:`scipy.optimize.curve_fit` to estimate the focus position
        and waist given arrays of axial `positions` and measured `radii`.

        Args:
            positions: Axial positions where radii were measured.
            radii: Measured beam radii corresponding to the positions.
            wavelength: Wavelength used to relate waist and Rayleigh range.
            p0: Initial guess for (focus, waist). If omitted a simple heuristic is used.

        Returns:
            Beam instance fitted to the data.
        """
        if p0 is None:  # TODO is this a good idea?
            p0 = (positions[np.argmin(radii)], np.min(radii))
        # yes using the class itself is pretty inefficient but its convenient and not performance critical
        (focus, waist), cov = curve_fit(
            lambda z, f, w: cls.from_gauss(f, w, wavelength).radius(z), positions, radii, p0=p0
        )
        return cls.from_gauss(focus, waist, wavelength, cov=cov)

    def plot(self, **kwargs) -> OpticalSetupPlot:  # pyright: ignore[reportPrivateImportUsage]
        """Plot the beam as part of an optical setup with no other elements.

        Args:
            **kwargs: Keyword arguments forwarded to :func:`OpticalSetup.plot`.

        Returns:
            OpticalSetupPlot instance for further customization.
        """

        return OpticalSetup(self, []).plot(**kwargs)

    def _repr_png_(self) -> bytes:
        fig, ax = plt.subplots()
        self.plot(ax=ax)
        return fig_to_png(fig)


# TODO general Element class?
@dataclass(frozen=True)
class ThinLens(YamlSerializableMixin):
    """Thin lens element including additional information."""

    focal_length: float  #: Focal length of the lens
    left_margin: float = 0.0  #: Physical size to the left of the focal plane
    right_margin: float = 0.0  #: Physical size to the right of the focal plane
    name: str | None = None  #: Name for reference and plotting

    def __post_init__(self):
        if self.focal_length == 0:
            raise ValueError("Focal length cannot be zero.")
        if self.left_margin + self.right_margin < 0:
            raise ValueError("Lens must have non-negative physical size.")  # focal plane outside physical lens is ok

    @cached_property
    def matrix(self) -> np.ndarray:
        """ABCD matrix of the lens element."""
        return np.array([[1, 0], [-1 / self.focal_length, 1]])

    def __str__(self) -> str:
        return self.name if self.name is not None else f"f={round(self.focal_length*1e3)}mm"


@dataclass(frozen=True)
class ThickLens(YamlSerializableMixin):
    """Thick lens element including additional information.

    The signs of the radius of curvature follow the standard optics convention where positive radii
    correspond to surfaces that are convex when viewed from the input side of the lens. This means
    that on the entry surface, a positive radius of curvature is a convex surface while on the exit surface,
    a negative radius of curvature is convex.
    Use :attr:`ThickLens.FLAT` to represent a flat surface. This is an alias for ``float('inf')``.
    """

    in_roc: float  #: Input surface radius of curvature, positive is convex
    out_roc: float  #: Output surface radius of curvature, negative is convex
    thickness: float  #: Thickness of the lens
    refractive_index: float  #: Refractive index of the lens material
    left_margin: float = 0.0  #: Physical size to the left of the lens center
    right_margin: float = 0.0  #: Physical size to the right of the lens center
    name: str | None = None  #: Name for reference and plotting

    FLAT: ClassVar[float] = float("inf")  #: Surface radius representing a flat surface

    def __post_init__(self):
        if self.thickness <= 0:
            raise ValueError("Lens thickness must be positive.")
        if self.refractive_index <= 1:
            raise ValueError("Refractive index must be greater than 1.")
        if self.left_margin + self.right_margin < 0:
            raise ValueError("Lens must have non-negative physical size.")

    @cached_property
    def matrix(self) -> np.ndarray:
        """ABCD matrix of the lens element."""
        n2 = self.refractive_index
        in_surface = np.array([[1, 0], [(1 - n2) / (self.in_roc * n2), 1 / n2]])
        propagation = np.array([[1, self.thickness], [0, 1]])
        out_surface = np.array([[1, 0], [(n2 - 1) / (self.out_roc), n2]])
        thickness_correction = np.array([[1, -self.thickness / 2], [0, 1]])
        return thickness_correction @ out_surface @ propagation @ in_surface @ thickness_correction

    @cached_property
    def focal_length(self) -> float:
        """Approximate focal length of the thick lens calculated using the
        `lensmaker's equation <https://en.wikipedia.org/wiki/Lens#Lensmaker's_equation>`_."""
        n2, r1, r2 = self.refractive_index, self.in_roc, self.out_roc
        if np.isinf(r1) and np.isinf(r2):
            return float("nan")
        return 1 / ((n2 - 1) * (1 / r1 - 1 / r2 + ((n2 - 1) * self.thickness) / (n2 * r1 * r2)))

    def __str__(self) -> str:
        return self.name if self.name is not None else f"f≈{round(self.focal_length*1e3)}mm"


Lens = ThinLens | ThickLens  #: Lens type union


@dataclass(frozen=True)
class OpticalSetup(YamlSerializableMixin):
    """Optical setup described by an initial beam and a sequence of elements.

    Implements :meth:`_repr_png_` to show a plot of the optical setup in IPython environments.
    """

    initial_beam: Beam  #: Initial beam before left most element
    elements: list[tuple[float, Lens]]  #: Optical elements as (position, element) tuples sorted by position
    validate: InitVar[bool] = True  #: Validate that elements are sorted by position

    def __post_init__(self, validate: bool) -> None:
        if validate and not all(left < right for (left, _), (right, _) in pairwise(self.elements)):
            raise ValueError("Optical elements must be sorted by position.")

    # TODO eliminate this and just put it into beams?
    @cached_property
    def beam_parameters(self) -> list[complex]:
        """Compute the ray vectors between elements including before the first element and after the last."""
        q = self.initial_beam.beam_parameter
        prev_pos = self.initial_beam.z_offset
        beam_parameters = [q]
        for pos, element in self.elements:
            q += pos - prev_pos  # free space propagation
            vec = element.matrix @ np.array([q, 1])  # lens transformation
            q = vec[0] / vec[1]  # normalize
            beam_parameters.append(q)
            prev_pos = pos
        return beam_parameters

    @cached_property
    def gauss_covariances(self) -> list[np.ndarray | None]:
        """Compute the covariance matrices between elements including before the first element and after the last."""
        if self.initial_beam.gauss_cov is None:
            return [None] * (len(self.elements) + 1)
        wavelength = self.initial_beam.wavelength
        cov = self.initial_beam.gauss_cov
        prev_pos = self.initial_beam.z_offset
        gauss_covariances = [cov]
        # for (pos, element), (q_prev, q_after) in zip(self.elements, pairwise(self.beam_parameters), strict=True):
        for (pos, element), (prev_beam, beam) in zip(self.elements, pairwise(self.beams_fast), strict=True):
            q_before = prev_beam.beam_parameter + (pos - prev_pos)  # beam parameter before element
            jac_fw_to_ri = self._jac_fw_to_ri(prev_beam.waist, wavelength)
            jac_ri_to_ri = self._jac_ri_to_ri(q_before, element.matrix)
            jac_ri_to_fw = self._jac_ri_to_fw(beam.beam_parameter, wavelength)
            jac_total = jac_ri_to_fw @ jac_ri_to_ri @ jac_fw_to_ri
            cov = jac_total @ cov @ jac_total.T
            gauss_covariances.append(cov)
            prev_pos = pos
        return gauss_covariances  # pyright: ignore[reportReturnType]

    @cached_property
    def beams_fast(self) -> list[Beam]:
        """Compute the Beam instances without propagating covariances."""
        return [self.initial_beam] + [
            Beam(beam_parameter=param, z_offset=pos, wavelength=self.initial_beam.wavelength)
            for (pos, _), param in zip(self.elements, self.beam_parameters[1:], strict=True)
        ]

    @cached_property
    def beams(self) -> list[Beam]:
        """Compute the Beam instances between elements including before the first element and after the last."""
        return [self.initial_beam] + [
            Beam(beam_parameter=param, z_offset=pos, wavelength=self.initial_beam.wavelength, gauss_cov=cov)
            for (pos, _), param, cov in zip(
                self.elements, self.beam_parameters[1:], self.gauss_covariances[1:], strict=True
            )
        ]

    # TODO optimize this to pass data to beams in batches?
    def radius(self, z: float | np.ndarray) -> float | np.ndarray:
        """Compute the beam radius at axial position(s)."""
        if not np.isscalar(z):
            return np.array([self.radius(zi) for zi in z])  # pyright: ignore[reportGeneralTypeIssues]
        index = np.searchsorted([pos for pos, _ in self.elements], z)
        return self.beams_fast[index].radius(z)  # pyright: ignore[reportArgumentType, reportCallIssue]

    # TODO only specify default k in one place?
    def radius_dev(self, z: float | np.ndarray) -> float | np.ndarray:
        """Compute the beam radius confidence interval at axial position(s)."""
        if not np.isscalar(z):
            return np.array([self.radius_dev(zi) for zi in z])  # pyright: ignore[reportGeneralTypeIssues]
        index = np.searchsorted([pos for pos, _ in self.elements], z)
        return self.beams[index].radius_dev(z)  # pyright: ignore[reportArgumentType, reportCallIssue]

    @classmethod
    def fit(
        cls,
        positions: np.ndarray,
        radii: np.ndarray,
        wavelength: float,
        elements: Sequence[tuple[float | tuple[float, float], Lens]],
        p0: tuple[float, float] | None = None,
    ) -> "OpticalSetup":
        """Fit an optical setup to measured data.

        Args:
            positions: Axial positions where radii were measured.
            radii: Measured beam radii corresponding to the positions.
            wavelength: Wavelength used to relate waist and Rayleigh range.
            elements: Optical elements as (position, element) tuples sorted by position.
                positions can be specified as float to describe a fixed position or as
                (expected, deviation) tuple to fit the element position between
                expected - deviation and expected + deviation.
            p0: Initial guess for (focus, waist). If omitted a simple heuristic is used.

        Returns:
            OpticalSetup instance fitted to the data.

        Raises:
            ValueError: If the initial guess for element positions are not in ascending order.
            ValueError: If the fitted element positions are invalid, i.e. not in ascending order.
            ValueError: If no data point is before the first element when using beam parameter initial guess heuristic.
        """
        fit_indices = [i for i, (pos, _) in enumerate(elements) if isinstance(pos, tuple)]

        def substitute_positions(elem_positions: Sequence[float]) -> list[tuple[float, Lens]]:
            substituted = list(elements)  # make a copy of potential non list sequence
            for idx, pos in zip(fit_indices, elem_positions, strict=True):
                substituted[idx] = (pos, substituted[idx][1])
            return substituted  # pyright: ignore[reportReturnType]

        def model(positions: np.ndarray, focus: float, waist: float, *elem_positions: float) -> np.ndarray:
            setup = cls(Beam.from_gauss(focus, waist, wavelength), substitute_positions(elem_positions), validate=False)
            return setup.radius(positions)  # pyright: ignore[reportReturnType]

        beam_guess = list(p0 if p0 is not None else (positions[np.argmin(radii)], np.min(radii)))
        initial_guess = beam_guess + [pos[0] for pos, _ in elements if isinstance(pos, tuple)]

        if not all(left < right for left, right in pairwise(initial_guess[2:])):
            raise ValueError("Initial guess for element positions must be in ascending order.")
        if p0 is None and np.all(positions >= initial_guess[2]):
            raise ValueError("At least one data point must be before the first element for initial guess heuristic.")

        element_bounds = [(pos[0] - pos[1], pos[0] + pos[1]) for pos, _ in elements if isinstance(pos, tuple)]
        bounds = np.transpose([(-np.inf, np.inf)] * 2 + element_bounds)
        params, cov = curve_fit(model, positions, radii, p0=initial_guess, bounds=bounds)
        focus, waist, *fitted_positions = params

        # warn if fitted positions are at bounds since this indicates a bad fit or too tight bounds
        for i, pos, (lower, upper) in zip(fit_indices, fitted_positions, element_bounds, strict=True):
            if abs(pos - lower) < 1e-6:
                warnings.warn(f"Fitted position of element {i} is at lower bound {lower}.", stacklevel=2)
            if abs(pos - upper) < 1e-6:
                warnings.warn(f"Fitted position of element {i} is at upper bound {upper}.", stacklevel=2)

        beam = Beam.from_gauss(focus, waist, wavelength, cov=cov[:2, :2])
        try:
            setup = cls(initial_beam=beam, elements=substitute_positions(fitted_positions))
        except ValueError as e:
            raise ValueError("Invalid fit result: optical elements are not sorted by position.") from e
        return setup

    plot = plot_setup  #: Plot the optical setup, see :func:`corset.plot.plot_setup`

    # TODO cache this (and the other repr png functions)?
    def _repr_png_(self) -> bytes:
        fig, ax = plt.subplots()
        self.plot(ax=ax)
        return fig_to_png(fig)

    # TODO include derivation code
    @staticmethod
    def _jac_fw_to_ri(waist: float, wavelength: float) -> np.ndarray:
        """Jacobian from [waist, wavelength] to [Re(q0p), Im(q0p)]
        ie from Gaussian beam parameters to complex beam parameter which
        has already been propagated to the new position.

        Args:
            waist: Beam waist radius.
            wavelength: Wavelength of the beam.

        Returns:
            2x2 Jacobian matrix.
        """
        # derivation:
        """
        >>> import sympy as sp
        >>> w0, f0, wl, z0, z1 = sp.symbols('waist focus wavelength z_{0} z_{1}', real=True)
        >>> q0 = z0 - f0 + sp.I * sp.pi * w0**2 / wl
        >>> q0p = q0 + z1 - z0 # propagate to just before element
        >>> print(sp.Matrix([sp.re(q0p), sp.im(q0p)]).jacobian(sp.Matrix([f0, w0])))
        Matrix([[-1, 0], [0, 2*pi*waist/wavelength]])
        """
        return np.array([[-1, 0], [0, 2 * np.pi * waist / wavelength]])

    @staticmethod
    def _jac_ri_to_ri(q0: complex, mat: np.ndarray) -> np.ndarray:
        """Jacobian from [Re(q0), Im(q0)] to [Re(q1), Im(q1)]
        ie from complex beam parameter before an optical element
        to complex beam parameter after the element.

        Args:
            q0: Complex beam parameter before the element.
            mat: 2x2 ABCD matrix of the optical element.

        Returns:
            2x2 Jacobian matrix.
        """
        # derivation:
        """
        >>> import sympy as sp
        >>> q0r, q0i, A, B, C, D = sp.symbols('q0.real q0.imag A B C D', real=True)
        >>> q0 = q0r + sp.I * q0i
        >>> q1 = ((A * q0 + B) / (C * q0 + D)).simplify()
        >>> jac = sp.simplify(sp.Matrix([sp.re(q1), sp.im(q1)]).jacobian(sp.Matrix([q0r, q0i])))
        >>> print("jac[0,0].factor())
        -(A*D - B*C)*(C*q0.imag - C*q0.real - D)*(C*q0.imag + C*q0.real + D)/(C**2*q0.imag**2 + C**2*q0.real**2 + 2*C*D*q0.real + D**2)**2
        >>> print("jac[0,1].factor())
        2*C*q0.imag*(A*D - B*C)*(C*q0.real + D)/(C**2*q0.imag**2 + C**2*q0.real**2 + 2*C*D*q0.real + D**2)**2
        >>> assert jac[0,0].factor() == jac[1,1].factor()
        >>> assert jac[1,0].factor() == -jac[0,1].factor()
        """
        A, B, C, D = mat.flat  # noqa: N806
        res = np.zeros((2, 2))
        # fmt: off
        res[0,0] = -(A*D - B*C)*(C*q0.imag - C*q0.real - D)*(C*q0.imag + C*q0.real + D)/(C**2*q0.imag**2 + C**2*q0.real**2 + 2*C*D*q0.real + D**2)**2
        res[0,1] = 2*C*q0.imag*(A*D - B*C)*(C*q0.real + D)/(C**2*q0.imag**2 + C**2*q0.real**2 + 2*C*D*q0.real + D**2)**2
        # fmt: on
        res[1, 0] = -res[0, 1]
        res[1, 1] = res[0, 0]
        return res

    @staticmethod
    def _jac_ri_to_fw(q1: complex, wavelength: float) -> np.ndarray:
        """Jacobian from [Re(q1), Im(q1)] to [waist, wavelength]
        ie from complex beam parameter after an optical element
        to Gaussian beam parameters.

        Args:
            q1: Complex beam parameter after the element.
            wavelength: Wavelength of the beam.

        Returns:
            2x2 Jacobian matrix.
        """
        # derivation:
        """
        >>> import sympy as sp
        >>> q1r, q1i, z1, wl = sp.symbols('q1.real q1.imag z_{1} wavelength', real=True, positive=True)
        >>> f1 = z1 - q1r
        >>> w1 = sp.sqrt(q1i * wl / sp.pi)
        >>> print(sp.Matrix([f1, w1]).jacobian(sp.Matrix([q1r, q1i])))
        Matrix([[-1, 0], [0, sqrt(wavelength)/(2*sqrt(pi)*sqrt(q1.imag))]])
        """
        return np.array([[-1, 0], [0, np.sqrt(wavelength) / (2 * np.sqrt(np.pi) * np.sqrt(q1.imag))]])



"""Configuration for analysis and plotting functions."""

# TODO is it a reasonable idea to just use this as a namespace?
# ideally these would just be values in this module but that would lead to bind by value issues when importing
import typing
from enum import Enum

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import to_rgb

T = typing.TypeVar("T")


class Unit(typing.NamedTuple):
    """Unit representation with ASCII and LaTeX strings and conversion factor to base unit."""

    ascii: str  #: ASCII representation of the unit
    tex: str  #: LaTeX representation of the unit
    factor: float  #: Conversion factor to the SI base unit


class Config:
    """Configuration namespace for the mode matching solver."""

    class SensitivityUnit(Enum):
        """Units for sensitivity analysis."""

        PER_M2 = Unit(ascii="%/m^2", tex=r"\%/\mathrm{m}^2", factor=1)  #:
        PERCENT_PER_MM2 = Unit(ascii="%/mm^2", tex=r"\%/\mathrm{mm}^2", factor=1e2 * (1e-3**2))  #:
        PERCENT_PER_CM2 = Unit(ascii="%/cm^2", tex=r"\%/\mathrm{cm}^2", factor=1e2 * (1e-2**2))  #:

    sensitivity_unit = SensitivityUnit.PERCENT_PER_CM2  #: Unit for sensitivity analyses

    overwrite_dark_theme: bool | None = None  #: Override automatic detection of dark theme in plots

    @classmethod
    def mpl_is_dark(cls) -> bool:
        """Determine whether the current Matplotlib theme is
        dark by analyzing the figure background color.

        Returns:
            `True` if the current Matplotlib theme is dark, `False` otherwise.
        """
        if cls.overwrite_dark_theme is not None:
            return cls.overwrite_dark_theme
        bg_color = to_rgb(plt.rcParams["figure.facecolor"])
        return bool(np.mean(bg_color[:3]) < 0.5)

    class Overlap:
        """Configuration of default values for overlap contour plots."""

        levels: typing.ClassVar[list[float]] = [80, 90, 95, 98, 99, 99.5, 99.8, 99.9, 100]
        """Overlap levels in percent."""
        colormap: str = "turbo"
        """Colormap for overlap levels."""
        grid_resolution: int = 50
        """Grid resolution for overlap contour plots."""

        @classmethod
        def colors(cls) -> typing.Sequence[tuple[float, float, float, float]]:
            """Get the colors corresponding to the overlap levels.

            Returns:
                A sequence of RGBA colors for each overlap level.
            """
            return plt.get_cmap(cls.colormap)(np.linspace(0, 1, len(cls.levels) - 1))  # type: ignore  # noqa: PGH003

    class PlotSetup:
        """Configuration of default values for setup plots."""

        rayleigh_range_cap: float = 200e-3
        """During plotting, the interval of interest is automatically determined
        based on the Rayleigh range of the beam, for large Rayleigh ranges this can
        significantly inflate the plotted region, making the important features hard to see.
        This parameter limits this effect by capping the maximum Rayleigh range considered when
        determining the plotted interval.
        """
        beam_points: int = 500
        """Number of different :math:`z` points to use when plotting beam propagation."""
        show_legend: bool = False
        """Whether to show a legend describing the plot asspects."""
        legend_loc: str = "lower left"
        """Location of the legend in the plot."""
        beam_kwargs: typing.ClassVar[dict] = {"color": "C0", "alpha": 0.5}
        """The kwargs to pass to :meth:`~matplotlib.axes.Axes.fill_between`"""
        confidence_interval: float | bool = 0.95
        """Confidence interval for the beam envelope. If `False`, no confidence interval is shown."""

    class PlotReachability:
        """Configuration of default values for reachability plots."""

        displacement: float | list[float] = 5e-3
        """Displacement(s) to use for all or each dimension (if a list)."""
        num_samples: int | list[int] = 7
        """Number of samples to use for all or each dimension (if a list)."""
        line_step: int | list[int] = 2
        """Line step to use for all or each dimension (if a list),
        i.e. the step size to skip certain lines for increased smoothness while retaining clarity."""
        confidence_interval: float | bool = 0.95
        """Confidence ellipse for focus position and waist radius. If `False`, no confidence ellipse is shown."""

    class PlotSensitivity:
        """Configuration of default values for sensitivity plots."""

        worst_overlap: float = 0.98
        """Approximate worst overlap that should still be fully contained in plot used for
        automatic plotting range determination."""
        num_samples_z: int = 3  # more is too cluttered most times
        """Number of z-slices to plot if 3 dimensions are used."""
        force_contour_lines: bool = False
        """Whether to always use contour lines instead of filled
        contours for plots with two degrees of freedom."""
        confidence_interval: float | bool = 0.95
        """Confidence ellipse probability in terms of the required x and y displacements.
        If `False`, no confidence ellipse is shown."""

    class PlotAll:
        """Configuration of default values for mode matching overview plots containing the setup,
        reachability, and sensitivity plots."""

        figsize: tuple[float, float] = (16, 4)
        """Figure size for the overview plot."""
        width_ratios: tuple[float, float, float] = (2, 1, 1)
        """Width ratios for the three subplots in the overview plot."""
        tight_layout: bool = True
        """Whether to use tight layout for the overview plot."""
        setup_kwargs: typing.ClassVar[dict] = {}
        """Keyword arguments to pass to the setup plot."""
        reachability_kwargs: typing.ClassVar[dict] = {}
        """Keyword arguments to pass to the reachability plot."""
        sensitivity_kwargs: typing.ClassVar[dict] = {}
        """Keyword arguments to pass to the sensitivity plot."""

    @staticmethod
    def get(value: T | None, default: T) -> T:
        """Get a configuration value, using the default if the provided value is `None`.

        Args:
            value: The value to check.
            default: The default value to use if `value` is `None`.

        Returns:
            The provided value or the default.
        """
        return default if value is None else value



from corset.solver import Lens, Region, lens_combinations


def _regions_repr(setup):
    """Represent the generated setup as tuple-of-tuples of focal lengths for easy assertions."""
    return tuple(tuple(lens.focal_length for lens in region) for region in setup)


class TestLensCombinations:
    def test_empty_regions_min_zero_yields_empty_setup(self):
        out = list(lens_combinations([], [], 0, 0))
        # Should yield a single empty setup (no regions)
        assert out == [[]]

    def test_empty_regions_with_positive_min_yields_nothing(self):
        out = list(lens_combinations([], [], 1, 2))
        assert out == []

    def test_single_region_uses_base_selection_when_selection_empty(self):
        base = [Lens(1.0, 0.0, 0.0), Lens(2.0, 0.0, 0.0)]
        reg = Region(left=0.0, right=1.0, min_elements=1, max_elements=2, selection=[])
        out = list(lens_combinations([reg], base, 1, 2))

        # Expect combinations_with_replacement of base of sizes 1 and 2
        expected = {((1.0,),), ((2.0,),), ((1.0, 1.0),), ((1.0, 2.0),), ((2.0, 2.0),)}
        got = set(_regions_repr(s) for s in out)
        assert got == expected

    def test_multiple_regions_and_global_limits(self):
        A = Lens(10.0, 0.0, 0.0)
        B = Lens(20.0, 0.0, 0.0)
        C = Lens(30.0, 0.0, 0.0)

        # First region: exactly one element chosen from {A,B}
        r1 = Region(left=0.0, right=1.0, min_elements=1, max_elements=1, selection=[A, B])
        # Second region: 0..2 elements chosen from {C}
        r2 = Region(left=1.0, right=2.0, min_elements=0, max_elements=2, selection=[C])

        out = list(lens_combinations([r1, r2], [], 1, 3))

        # Build expected: r1 must be (10,) or (20,); r2 can be (), (30,), (30,30)
        expected = set()
        for first in ((10.0,), (20.0,)):
            for second in ((), (30.0,), (30.0, 30.0)):
                expected.add((first, second))

        got = set(_regions_repr(s) for s in out)
        assert got == expected

    def test_region_min_greater_than_global_max_yields_nothing(self):
        A = Lens(1.0, 0.0, 0.0)
        r = Region(left=0.0, right=1.0, min_elements=3, max_elements=3, selection=[A])
        out = list(lens_combinations([r], [], 0, 2))
        assert out == []



# type: ignore  # noqa: PGH003
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np

from corset.core import Beam

rainbow = [
    "#cd001a",
    "#ef6a00",
    "#f2cd00",
    "#79c300",
    "#1961ae",
    "#61007d",
]
corset = ("#1a1a1a", "#e6e6e6")
lace = ("#cccccc", "#333333")


def make_logo(filename, dark, crop=None):
    # setup plot
    figsize = (8, 4.5) if not crop else (crop[0][1] - crop[0][0], crop[1][1] - crop[1][0])
    fig, ax = plt.subplots(figsize=figsize, tight_layout=True)
    ax.set_aspect("equal")
    ax.set_xlim(-8, 8)
    ax.set_ylim(-4.5, 4.5)
    ax.axis("off")
    fig.set_facecolor("none")

    beam = Beam.from_gauss(focus=0, waist=1, wavelength=1)
    radius_scale = 1.6

    # make the beam
    zs = np.linspace(-8, 8, 100)
    rs = beam.radius(zs)
    fractions = np.linspace(-1, 1, len(rainbow) + 1) * radius_scale
    half = len(zs) // 2
    eps = -5e-2
    for c1, c2, lower, upper in zip(rainbow, rainbow[::-1], fractions[:-1], fractions[1:], strict=True):
        ax.fill_between(zs[:half], rs[:half] * lower - eps, rs[:half] * upper + eps, color=c1, ec="none")
        ax.fill_between(zs[half:], rs[half:] * lower - eps, rs[half:] * upper + eps, color=c2, ec="none")

    # make the corset
    corset_color = corset[1] if dark else corset[0]
    corset_left, corset_right = -2.5, 2.5
    zs_corset = np.linspace(corset_left, corset_right, 50)
    rs_corset = beam.radius(zs_corset) * radius_scale
    ax.fill_between(zs_corset, -rs_corset, rs_corset, color=corset_color, lw=15)

    # make the laces
    lace_color = lace[1] if dark else lace[0]
    n_laces = 3
    bounds = np.linspace(corset_left, corset_right, n_laces + 1)
    width = np.diff(bounds)[0] * 0.6
    positions = (bounds[:-1] + bounds[1:]) / 2
    for pos in positions:
        kwargs = {"color": lace_color, "lw": 6, "solid_capstyle": "round"}
        ax.plot([pos - width / 2, pos + width / 2], [-width / 2, width / 2], **kwargs)
        ax.plot([pos - width / 2, pos + width / 2], [width / 2, -width / 2], **kwargs)

    if crop is not None:
        ax.set_xlim(*crop[0])
        ax.set_ylim(*crop[1])

    fig.savefig(filename, transparent=True)


if __name__ == "__main__":
    # run with "pixi run -e dev python ./misc/logo/gen_logo.py"
    directory = Path(__file__).parent.resolve()
    make_logo(directory / "logo_light.svg", dark=False)
    make_logo(directory / "logo_dark.svg", dark=True)
    # only show the right expanding part for favicon
    make_logo(directory / "favicon.svg", dark=False, crop=([0, 8], [-4, 4]))

    favicon_svg = (directory / "favicon.svg").read_text()
    favicon_svg = favicon_svg.replace(corset[0], "var(--corset)").replace(lace[0], "var(--lace)")
    favicon_svg = favicon_svg.replace(
        "</style>",
        f"\n:root {{ --corset: {corset[0]}; --lace: {lace[0]}; }}\n@media (prefers-color-scheme: dark) {{ :root {{ --corset: {corset[1]}; --lace: {lace[1]}; }} }}\n</style>",
    )
    (directory / "favicon.svg").write_text(favicon_svg)



"""Mode matching solver and related data structures.

The main entry point is the function :func:`mode_match`, it will setup
the :class:`ModeMatchingProblem`.
The problem defined with the help of :class:`ShiftingRange`, :class:`Aperture`, and :class:`Passage` instances.
The constructed :class:`ModeMatchingProblem` yields :class:`ModeMatchingCandidate` instances
that are then optimized to produce :class:`ModeMatchingSolution` instances.
Each step containing a reference to the data structure it is based on.
The actual constrained optimization is carried out using :func:`scipy.optimize.minimize` using the SLSQP method.
All solutions will then be collected in a :class:`SolutionList` for convenient analysis and filtering.
"""

from collections.abc import Callable, Generator, Iterable, Iterator, Sequence
from dataclasses import dataclass, field
from functools import cached_property
from itertools import combinations_with_replacement, pairwise
from typing import Any, cast, overload

import numpy as np
import pandas as pd
from scipy import optimize

from .analysis import ModeMatchingAnalysis
from .core import Beam, Lens, OpticalSetup
from .plot import (
    fig_to_png,
    plot_mode_match_solution_all,
    plot_mode_match_solution_setup,
    plot_reachability,
    plot_sensitivity,
)
from .serialize import YamlSerializableMixin


@dataclass(frozen=True)
class ShiftingRange(YamlSerializableMixin):
    """Range where lenses can be placed."""

    left: float  #: Range left boundary
    right: float  #: Range right boundary
    min_elements: int = 0  #: Minimum number of lenses to be placed in this range
    max_elements: int = float("inf")  # pyright: ignore[reportAssignmentType]
    """Maximum number of lenses that can be placed in this range."""
    selection: Iterable[Lens] = field(default_factory=list)
    """Optional set of lenses to use for this range, if empty the global selection is used."""

    def __post_init__(self):
        if self.right <= self.left:
            raise ValueError("Range right boundary must be greater than left boundary.")
        if self.min_elements < 0:
            raise ValueError("min_elements cannot be negative.")
        if self.max_elements < self.min_elements:
            raise ValueError("max_elements cannot be less than min_elements.")


@dataclass(frozen=True)
class ParametrizedSetup(YamlSerializableMixin):
    """Parametrized optical setup where some or all element positions are free parameters."""

    initial_beam: Beam  #: Initial beam before left most element
    elements: list[tuple[float | None, Lens]]
    """Optical elements as (position | None, element) tuples. None represents a free element / parameter."""

    def substitute(self, positions: list[float] | np.ndarray, validate: bool = True) -> OpticalSetup:
        """Substitute free parameters with given positions.

        Args:
            positions: Positions to substitute for free parameters in order, must match number of free parameters.
            validate: Whether to validate the positions in the resulting setup.

        Returns:
            The resulting optical setup with substituted positions.

        Raises:
            ValueError: If the number of provided positions does not match the number of free parameters.

        """
        substituted_elements = []
        pos_index = 0
        try:
            for pos, element in self.elements:
                if pos is None:
                    pos = positions[pos_index]
                    pos_index += 1
                substituted_elements.append((pos, element))
        except IndexError as e:
            raise ValueError("Not enough positions provided to substitute all parametrized elements.") from e
        if pos_index < len(positions):
            raise ValueError("Too many positions provided for the number of parametrized elements.")
        return OpticalSetup(self.initial_beam, substituted_elements, validate)

    @cached_property
    def free_elements(self) -> list[int]:
        """List of indices of free elements (positions that are None)."""
        return [i for i, (pos, _) in enumerate(self.elements) if pos is None]


@dataclass(frozen=True)
class Aperture(YamlSerializableMixin):
    """Aperture constraint for a mode matching problem."""

    position: float  #: Axial position of the aperture
    radius: float  #: Aperture radius
    power: float = float(1 - np.exp(-2))
    r"""Minimum fraction of power that must pass through the aperture.
    If set to :math:`1 - e^{-2} \approx 0.865`, the beam's :math:`1/e` amplitude
    radius at the aperture is constrained to be smaller than the aperture radius."""

    def __post_init__(self):
        if self.radius <= 0:
            raise ValueError("Aperture radius must be positive.")

    @cached_property
    def radius_constraints(self) -> list[tuple[float, float]]:
        """A list containing this aperture as a normalized :math:`1/e` beam radius constraint
        tuple (position, radius). This is a tuple of a tuple to provide a uniform interface
        with :class:`Passage` which is a tuple of two tuple.
        """
        return [(self.position, np.sqrt(-2 / np.log(1 - self.power)) * self.radius)]


@dataclass(frozen=True)
class Passage(YamlSerializableMixin):
    """Passage constraint for a mode matching problem.
    A passage from `left` to `right` with given `radius` is represented (i.e. equivalent)
    to two :class:`Aperture` constraints with `radius` at the `left` and `right` boundaries.
    """

    left: float  #: Left boundary of the passage
    right: float  #: Right boundary of the passage
    radius: float  #: Passage radius
    power: float = float(1 - np.exp(-2))
    r"""Minimum fraction of power that must pass through the passage (approximately).
    Since the passage is represented as two apertures, this is really just the power
    fraction of the respective apertures. If set to :math:`1 - e^{-2} \approx 0.865`,
    the beam's :math:`1/e` amplitude radius at the apertures is constrained to be smaller
    than the passage radius. See :attr:`Aperture.power`.
    """

    def __post_init__(self):
        if self.right <= self.left:
            raise ValueError("Left boundary must be less than right boundary.")
        if self.radius <= 0:
            raise ValueError("Passage radius must be positive.")

    @classmethod
    def centered(cls, center: float, width: float, radius: float, power: float = float(1 - np.exp(-2))) -> "Passage":
        """Convenience constructor for a passage centered at a given position.

        Args:
            center: Center position of the passage.
            width: Width of the passage.
            radius: Passage radius.
            power: Minimum power fraction that must pass through the passage.


        Returns:
            Passage instance centered at the given position.
        """
        half_width = width / 2
        return cls(left=center - half_width, right=center + half_width, radius=radius, power=power)

    @cached_property
    def radius_constraints(self) -> list[tuple[float, float]]:
        """Representation of the constraint as two normalized :math:`1/e` beam radius constraints.
        A list of two tuples (position, radius) for the left and right boundaries.
        """
        normalized_radius = np.sqrt(-2 / np.log(1 - self.power)) * self.radius
        return [(self.left, normalized_radius), (self.right, normalized_radius)]


# TODO should this be a member of Beam?
def mode_overlap(delta_z: float, waist_a: float, waist_b: float, wavelength: float) -> float:
    """Compute the mode overlap between two Gaussian beams.

    Args:
        delta_z: Axial distance between the beam waists.
        waist_a: Waist radius of the first beam.
        waist_b: Waist radius of the second beam.
        wavelength: Wavelength of the beams.

    Returns:
        Mode overlap between the two beams.
    """
    return (
        2 * np.pi * waist_a * waist_b / np.sqrt(wavelength**2 * delta_z**2 + np.pi**2 * (waist_a**2 + waist_b**2) ** 2)
    )


@dataclass(frozen=True)
class ModeMatchingProblem(YamlSerializableMixin):
    """Mode matching problem definition."""

    setup: OpticalSetup  #: Initial optical setup
    desired_beam: Beam  #: Desired output beam after the optical setup
    ranges: list[ShiftingRange]  #: Ranges where lenses can be placed
    selection: Sequence[Lens]  #: Selection of lenses to choose from for mode matching
    min_elements: int  #: Minimum number of elements to use
    max_elements: int  #: Maximum number of elements to use
    constraints: Sequence[Aperture | Passage]  #: Beam constraints on the optical setup

    # TODO maybe also verify that there is a combination of lenses that can fit in the ranges?
    def __post_init__(self):
        self._verify_selection()
        self._verify_no_overlaps()
        if self.setup.initial_beam.wavelength != self.desired_beam.wavelength:
            # testing for equality of floats should be fine here since they should come from the same source
            raise ValueError("Setup initial beam and desired beam must have the same wavelength.")

    def _verify_selection(self):
        if self.min_elements < 1:
            raise ValueError("Global min_elements must be at least 1.")

        if self.max_elements is not None and self.max_elements < self.min_elements:
            raise ValueError("Global max_elements cannot be less than min_elements.")

        total_min = sum(r.min_elements for r in self.ranges)
        total_max = sum(r.max_elements for r in self.ranges)

        if total_max == float("inf") and self.max_elements == float("inf"):
            raise ValueError("Cannot have unbounded maximum elements when global maximum is not set.")

        if total_min > self.max_elements:
            raise ValueError("Sum of range minimum elements exceeds global maximum elements.")
        if total_max < self.min_elements:
            raise ValueError("Sum of range maximum elements is less than global minimum elements.")

    def _verify_no_overlaps(self):
        regions: list[tuple[float, float, Any]] = []  # left right
        regions.extend((pos, pos, element) for pos, element in self.setup.elements)
        regions.extend((range_.left, range_.right, range_) for range_ in self.ranges)
        regions.extend((ap.position, ap.position, ap) for ap in self.radius_constraints if isinstance(ap, Aperture))
        regions.extend((pas.left, pas.right, pas) for pas in self.constraints if isinstance(pas, Passage))
        regions.sort(key=lambda x: x[0])
        for r1, r2 in pairwise(regions):
            if r1[1] > r2[0]:
                raise ValueError(f"Overlapping regions/elements detected: {r1[2]} and {r2[2]}.")

    @cached_property
    def radius_constraints(self) -> list[tuple[float, float]]:
        """List of all constraints as :math:`1/e` beam radius constraint tuples (position, radius)."""
        return [con for constraint in self.constraints for con in constraint.radius_constraints]

    @cached_property
    def interleaved_elements(self) -> list[tuple[float, Lens] | int]:
        """List of all (potential) elements, ranges where elements can be placed are represented
        by their index instead of the (position, element) tuple. Used for easy construction of parametrized setups."""
        merged: list[tuple[float, Lens] | int] = []
        next_boundary = self.ranges[0].left if self.ranges else float("inf")
        range_index = 0

        for pos, element in self.setup.elements:
            while pos > next_boundary:
                merged.append(range_index)
                range_index += 1
                next_boundary = self.ranges[range_index].left if range_index < len(self.ranges) else float("inf")
            merged.append((pos, element))

        while range_index < len(self.ranges):
            merged.append(range_index)
            range_index += 1

        return merged

    @classmethod
    def lens_combinations(
        cls,
        ranges: list[ShiftingRange],
        base_selection: Sequence[Lens],
        min_elements: int,
        max_elements: int,
        current_populations: list[tuple[Lens, ...]] = [],  # noqa: B006
    ) -> Generator[list[tuple[Lens, ...]], None, None]:
        """Recursive helper function to generate all possible lens populations for the given ranges.

        Args:
            ranges: Remaining ranges to process.
            base_selection: Global selection of lenses to choose from.
            min_elements: Minimum number of elements remaining to place.
            max_elements: Maximum number of elements remaining to place.
            current_populations: Current lens populations for the processed ranges.

        Yields:
            list[tuple[Lens, ...]]: The next lens populations for all ranges.
        """

        if not ranges:
            if min_elements <= 0:
                yield current_populations
            return

        first, *rest = ranges

        if first.min_elements > max_elements:
            return

        for num_elements in range(first.min_elements, min(first.max_elements, max_elements) + 1):
            selection = first.selection if first.selection else base_selection
            for comb in combinations_with_replacement(selection, num_elements):
                if sum(lens.left_margin + lens.right_margin for lens in comb) > (first.right - first.left):
                    continue  # skip combinations that cannot fit in the range
                new_setup = [*current_populations, comb]
                yield from cls.lens_combinations(
                    rest, base_selection, min_elements - num_elements, max_elements - num_elements, new_setup
                )

    def candidates(self) -> Generator["ModeMatchingCandidate", None, None]:
        """Generate all possible range population candidates for the mode matching problem.

        Returns:
            Generator of :class:`ModeMatchingCandidate` instances.
        """
        for population in self.lens_combinations(self.ranges, self.selection, self.min_elements, self.max_elements):
            yield ModeMatchingCandidate(problem=self, populations=population)


@dataclass(frozen=True)
class ModeMatchingCandidate(YamlSerializableMixin):
    """A candidate lens population for a mode matching problem."""

    problem: ModeMatchingProblem  #: The parent mode matching problem
    populations: list[tuple[Lens, ...]]  #: Lens populations for each range

    def generate_initial_positions(
        self, randomize: bool, rng: np.random.Generator = np.random.default_rng()  # noqa: B008
    ) -> np.ndarray:
        """Generate a set of initial positions for free lenses of the candidate.

        Args:
            randomize: Whether to generate random initial positions or evenly spaced positions.
            rng: Random number generator to use for random positions.

        Returns:
            Array of initial positions for the lenses.

        Raises:
            ValueError: If there is not enough space in a range for the lenses with their margins.
        """
        positions = []
        for range_, population in zip(self.problem.ranges, self.populations, strict=True):
            if not population:
                continue
            total_margin = sum(lens.left_margin + lens.right_margin for lens in population)
            available_space = range_.right - range_.left - total_margin
            if available_space < 0:
                raise ValueError("Not enough space in range for the lenses with their margins.")

            if randomize:
                distances = np.diff(np.sort(rng.uniform(0, available_space, len(population))), prepend=0)
            else:
                distances = np.repeat(available_space / (len(population) + 1), len(population))

            current_pos = range_.left
            for lens, distance in zip(population, distances, strict=True):
                current_pos += lens.left_margin + distance
                positions.append(current_pos)
                current_pos += lens.right_margin

        return np.array(positions)

    @cached_property
    def parametrized_setup(self) -> ParametrizedSetup:
        """Parametrized optical setup for the candidate."""
        elements = []
        for elem in self.problem.interleaved_elements:
            if isinstance(elem, int):
                for lens in self.populations[elem]:
                    elements.append((None, lens))
            else:
                elements.append(elem)
        return ParametrizedSetup(self.problem.setup.initial_beam, elements)

    @cached_property
    def position_constraint(self) -> optimize.LinearConstraint:
        """Linear constraint ensuring lenses do not overlap, stay within ranges and do not swap order."""
        constraints: list[tuple[np.ndarray, float, float]] = []
        pop_sizes = [len(pop) for pop in self.populations]
        index_offsets = np.cumsum([0, *pop_sizes[:-1]])
        mask = np.identity(sum(pop_sizes))
        for range_, population, base_idx in zip(self.problem.ranges, self.populations, index_offsets, strict=True):
            if not population:
                continue
            if len(population) == 1:
                constraints.append(
                    (mask[base_idx], range_.left + population[0].left_margin, range_.right - population[0].right_margin)
                )
            else:
                right_index = base_idx + len(population) - 1
                constraints.append((mask[base_idx], range_.left + population[0].left_margin, np.inf))
                constraints.append((mask[right_index], -np.inf, range_.right - population[-1].right_margin))

            for i in range(len(population) - 1):
                left_lens = population[i]
                right_lens = population[i + 1]
                constraints.append(
                    (
                        mask[base_idx + i + 1] - mask[base_idx + i],
                        left_lens.right_margin + right_lens.left_margin,
                        np.inf,
                    )
                )
        cols, lb, ub = zip(*constraints, strict=True)
        return optimize.LinearConstraint(
            np.vstack(cols), np.array(lb), np.array(ub)  # pyright: ignore[reportArgumentType]
        )

    @cached_property
    def beam_constraint(self) -> optimize.NonlinearConstraint:
        """Nonlinear constraint to ensure beam radius is within aperture constraints."""
        zs, rs = np.transpose(self.problem.radius_constraints)
        return optimize.NonlinearConstraint(
            lambda x, zs=zs, rs=rs, s=self.parametrized_setup: s.substitute(x, validate=False).radius(zs) / rs, 0, 1
        )

    @cached_property
    def constraints(self) -> list[optimize.NonlinearConstraint | optimize.LinearConstraint]:
        """List of position constraints and beam constraints if there are any."""
        return [self.position_constraint] + ([self.beam_constraint] if self.problem.constraints else [])

    def parametrized_overlap(self, positions: np.ndarray) -> float:
        """Compute the mode overlap depending on the free element positions.

        Args:
            positions: Positions of the free elements.

        Returns:
            Mode overlap value.
        """
        setup = self.parametrized_setup.substitute(positions, validate=False)
        final_beam = Beam(
            setup.beam_parameters[-1], z_offset=setup.elements[-1][0], wavelength=setup.initial_beam.wavelength
        )
        return mode_overlap(
            final_beam.focus - self.problem.desired_beam.focus,
            final_beam.waist,
            self.problem.desired_beam.waist,
            self.problem.setup.initial_beam.wavelength,
        )

    def parametrized_focus_and_waist(self, positions: np.ndarray) -> np.ndarray:
        """Compute the final beam focus and waist depending on the free element positions.

        Args:
            positions: Positions of the free elements.

        Returns:
            Array of final beam focus and waist.
        """
        setup = self.parametrized_setup.substitute(positions, validate=False)
        final_beam = setup.beams_fast[-1]
        return np.array([final_beam.focus, final_beam.waist])

    def optimize(
        self,
        filter_pred: Callable[["ModeMatchingSolution"], bool],
        random_initial_positions: int,
        equal_setup_tol: float,
        max_solutions: int,
        rng: np.random.Generator = np.random.default_rng(),  # noqa: B008
        # optimize_coupling: bool,
    ) -> list["ModeMatchingSolution"]:
        """Optimize the candidate to find mode matching solutions.

        Args:
            filter_pred: Predicate function that must return True for a solution to be accepted.
            random_initial_positions: Number of random initial positions to generate in addition to the equally spaced one.
            equal_setup_tol: Tolerance for considering two solutions as equal for eliminating duplicates.
            max_solutions: Maximum number of solutions to optimize and return.

        Returns:
            List of optimized mode matching solutions for the candidate.
        """
        # TODO make it so that the order of evaluating the candidates does not change their random values
        initial_setups = [self.generate_initial_positions(randomize=False, rng=rng)]
        for _ in range(random_initial_positions):
            initial_setups.append(self.generate_initial_positions(randomize=True, rng=rng))

        # TODO unify this with the make_mode_overlap function in analyze.py?
        def objective(positions: np.ndarray) -> float:
            setup = self.parametrized_setup.substitute(positions, validate=False)  # pyright: ignore[reportArgumentType]
            initial_beam = self.problem.setup.initial_beam
            final_beam = Beam(
                setup.beam_parameters[-1], z_offset=setup.elements[-1][0], wavelength=initial_beam.wavelength
            )
            desired_beam = self.problem.desired_beam
            return -mode_overlap(
                final_beam.focus - desired_beam.focus,
                final_beam.waist,
                desired_beam.waist,
                final_beam.wavelength,  # pyright: ignore[reportArgumentType]
            )

        solutions = []
        solution_positions = []  # for this lens population
        constrained_initial_setups = []
        for x0 in initial_setups:
            if len(solutions) >= max_solutions:
                break

            if self.problem.constraints:
                constrain_res = optimize.minimize(
                    lambda x: np.max(self.beam_constraint.fun(x)),
                    x0,
                    constraints=[self.position_constraint],
                    method="SLSQP",
                    options={"ftol": 2e-1},
                )
                if not np.all(self.beam_constraint.fun(constrain_res.x) <= self.beam_constraint.ub) or any(
                    not np.allclose(constrain_res.x, x0) for x0 in constrained_initial_setups
                ):
                    # failed to converge to a feasible setup or already tried this setup
                    continue
                constrained_initial_setups.append(constrain_res.x)
                x0 = constrain_res.x

            res = optimize.minimize(objective, x0, constraints=self.constraints)
            if not res.success:
                continue

            sol = ModeMatchingSolution(
                candidate=self,
                overlap=-res.fun,
                positions=res.x,
            )
            if any(np.allclose(sol.positions, pos, atol=equal_setup_tol, rtol=0) for pos in solution_positions):
                continue
            if filter_pred(sol):  # pyright: ignore[reportCallIssue]
                solutions.append(sol)
                solution_positions.append(sol.positions)

            # if optimize_coupling and sol.overlap >= 1 - 1e-3 and len(solutions) < solution_per_population:
            #     improved_sol = sol.optimize_coupling()
            #     if improved_sol is not None and filter_pred(improved_sol):  # pyright: ignore[reportCallIssue]
            #         if any(
            #             np.allclose(improved_sol.positions, pos, atol=equal_setup_tol, rtol=0)
            #             for pos in solution_positions
            #         ):
            #             continue
            #         solutions.append(improved_sol)
            #         solution_positions.append(improved_sol.positions)

        return solutions


@dataclass(frozen=True)
class ModeMatchingSolution(YamlSerializableMixin):
    """A solution to a mode matching problem.

    Implements :meth:`_repr_png_` to show a plot of the solution setup in IPython environments.
    """

    candidate: ModeMatchingCandidate  #: The candidate that produced this solution
    overlap: float  #: Mode overlap of the solution
    positions: np.ndarray  #: Positions of the lenses in the solution

    @property
    def setup(self) -> OpticalSetup:
        """Optical setup corresponding to this solution."""
        return self.candidate.parametrized_setup.substitute(self.positions)  # pyright: ignore[reportArgumentType]

    plot_setup = plot_mode_match_solution_setup  #: Plot the solution setup, see :func:`corset.plot.plot_mode_match_solution_setup`
    plot_reachability = plot_reachability  #: Plot the reachability analysis, see :func:`corset.plot.plot_reachability`
    plot_sensitivity = plot_sensitivity  #: Plot the sensitivity analysis, see :func:`corset.plot.plot_sensitivity`
    plot_all = plot_mode_match_solution_all  #: Plot all analyses, see :func:`corset.plot.plot_mode_match_solution_all`

    def _repr_png_(self) -> bytes:
        fig, _ = self.plot_all()
        return fig_to_png(fig)

    @cached_property
    def analysis(self) -> "ModeMatchingAnalysis":
        """Analysis data for this solution."""
        from . import analysis

        return analysis.ModeMatchingAnalysis(solution=self)

    def optimize_coupling(
        self, min_abs_improvement: float = 0.1, min_rel_improvement: float = 0.5
    ) -> "ModeMatchingSolution | None":
        """Optimize the solution to reduce coupling between least coupled pair while maintaining mode matching.

        Note that this requires at least 3 elements so that there is at least one degree of freedom left after mode matching.

        Args:
            min_abs_improvement: Minimum absolute improvement in coupling to return the new solution.
            min_rel_improvement: Minimum relative improvement in coupling to return the new solution.

        Returns:
            A new :class:`ModeMatchingSolution` with improved coupling if successful, otherwise None.

        Raises:
            ValueError: If there are not enough free parameters or if the solution does not have ~100% mode overlap.
        """

        if not len(self.positions) > 2:
            return None
            # raise ValueError("Need at least 3 free parameters to optimize coupling.")
        if self.overlap < 1 - 1e-3:
            raise ValueError("Can only optimize coupling for solutions ~100% mode overlap.")

        desired_beam = self.candidate.problem.desired_beam
        desired_focus_and_waist = np.array([desired_beam.focus, desired_beam.waist])
        mode_matching_constraint = optimize.NonlinearConstraint(
            lambda x: self.candidate.parametrized_focus_and_waist(x) - desired_focus_and_waist, [0, 0], [0, 0]
        )

        res = optimize.minimize(
            lambda x: ModeMatchingSolution(
                candidate=self.candidate,
                overlap=1,
                positions=x,
            ).analysis.min_coupling,
            self.positions,
            constraints=[*self.candidate.constraints, mode_matching_constraint],
        )

        if not res.success:
            return None

        mode_matching = self.candidate.parametrized_overlap(res.x)
        new_solution = ModeMatchingSolution(
            candidate=self.candidate,
            overlap=mode_matching,
            positions=res.x,
        )
        old_coupling = self.analysis.min_coupling
        new_coupling = new_solution.analysis.min_coupling
        if (old_coupling - new_coupling) > min_abs_improvement or new_coupling / old_coupling < min_rel_improvement:
            return new_solution
        return None


@dataclass(frozen=True)
class SolutionList(YamlSerializableMixin):
    """List of mode matching solutions with convenient methods for filtering and sorting.

    Supports (array) indexing, iteration, and other list-like operations.
    Implements :meth:`_repr_html_` to show a :class:`pandas.DataFrame` representation in IPython environments.
    """

    solutions: list[ModeMatchingSolution]  #: Underlying list of mode matching solutions

    @overload
    def __getitem__(self, index: int) -> "ModeMatchingSolution": ...

    @overload
    def __getitem__(self, index: slice | list[int]) -> "SolutionList": ...

    def __getitem__(self, index: int | slice | list[int]) -> "ModeMatchingSolution | SolutionList":
        if isinstance(index, int):
            return self.solutions[index]
        else:
            return SolutionList(solutions=np.array(self.solutions)[index].tolist())

    def __len__(self) -> int:
        return len(self.solutions)

    def __iter__(self) -> Iterator[ModeMatchingSolution]:
        return iter(self.solutions)

    @cached_property
    def df(self) -> pd.DataFrame:
        """DataFrame representation of the solutions for convenient analysis."""
        return pd.DataFrame([sol.analysis.summary() for sol in self.solutions])

    def _repr_html_(self) -> str:
        return self.df.to_html(notebook=True)

    def query(self, expr: str) -> "SolutionList":
        """Filter solutions based on a :meth:`pandas.DataFrame.query` expression applied to the DataFrame representation.

        Args:
            expr: Query string, see :meth:`pandas.DataFrame.query` for details.

        Returns:
            A new :class:`SolutionList` containing only the solutions that satisfy the query.
        """
        return self[cast(list[int], self.df.query(expr).index)]

    def sort_values(self, by: str | list[str], ascending: bool = True) -> "SolutionList":
        """Sort solutions based on a column or list of columns in the DataFrame representation.

        Args:
            by: Column name or list of column names to sort by.
            ascending: Whether to sort in ascending order.

        Returns:
            A new :class:`SolutionList` with solutions sorted by the specified columns.
        """
        return self[cast(list[int], self.df.sort_values(by=by, ascending=ascending).index)]


# TODO should this be a method of ModeMatchingProblem?
def mode_match(
    setup: Beam | OpticalSetup,
    desired_beam: Beam,
    ranges: list[ShiftingRange],
    selection: Sequence[Lens] = [],
    min_elements: int = 1,
    max_elements: int = float("inf"),  # pyright: ignore[reportArgumentType]
    constraints: Sequence[Aperture | Passage] = [],
    filter_pred: Callable[[ModeMatchingSolution], bool] | float | None = 0.999,  # allow for some numerical error
    random_initial_positions: int = 0,
    solutions_per_candidate: int = 1,
    equal_setup_tol: float = 1e-3,
    random_seed: int = 0,
    # optimize_coupling: bool = False,
    # pure_constraints: bool = False,  # TODO also give constraints a slight weight when there are excess degrees of freedom
    # TODO other solver options
):
    """Solve a mode matching problem by optimizing lens positions while respecting constraints.

    Args:
        setup: Initial optical setup or initial beam.
        desired_beam: Desired output beam after the optical setup.
        ranges: Shifting ranges where lenses can be placed.
        selection: Selection of lenses to choose from for mode matching.
        min_elements: Minimum number of elements to use.
        max_elements: Maximum number of elements to use.
        constraints: Beam constraints on the resulting optical setup.
        filter_pred: Predicate function or minimum overlap float to filter solutions.
        random_initial_positions: Number of random initial positions to generate per candidate.
        solutions_per_candidate: Maximum number of solutions to optimize and return per candidate i.e. lens population.
        equal_setup_tol: Tolerance for considering two solutions as equal for eliminating duplicates.
        random_seed: Random seed for reproducibility.

    Returns:
        A :class:`SolutionList` containing the optimized mode matching solutions.
    """

    # verify and prepare inputs
    if isinstance(setup, Beam):
        setup = OpticalSetup(setup, [])

    if isinstance(filter_pred, float):
        min_overlap = filter_pred
        filter_pred = lambda s: s.overlap >= min_overlap
    elif filter_pred is None:
        filter_pred = lambda s: True

    problem = ModeMatchingProblem(
        setup=setup,
        desired_beam=desired_beam,
        ranges=ranges,
        selection=selection,
        min_elements=min_elements,
        max_elements=max_elements,
        constraints=constraints,
    )

    rng = np.random.default_rng(random_seed)

    solutions = []
    # TODO parallelize this loop?
    for candidate in problem.candidates():
        solutions.extend(
            candidate.optimize(
                filter_pred=filter_pred,  # pyright: ignore[reportArgumentType]
                random_initial_positions=random_initial_positions,
                equal_setup_tol=equal_setup_tol,
                max_solutions=solutions_per_candidate,
                rng=rng,
                # optimize_coupling=optimize_coupling,
            )
        )

    return SolutionList(solutions)

    # TODO some sanity checks to ensure the desired beam is after the last setup part
    # TODO other checks like non overlapping regions
    # TODO return special solution list type that allows sorting and filtering and other convenient stuff



"""Analysis tools for detailed analysis of mode matching solutions."""

import warnings
from collections.abc import Callable, Iterable
from dataclasses import dataclass
from functools import cached_property, wraps
from typing import TYPE_CHECKING, cast

import numpy as np
import pandas as pd
from scipy.differentiate import hessian, jacobian

from .config import Config

if TYPE_CHECKING:
    from .solver import ModeMatchingSolution


def wrap_for_differentiate(
    func: Callable[[np.ndarray], np.ndarray],
) -> Callable[[np.ndarray], np.ndarray]:
    """Wrap a function to implement the vectorized input/output behavior
    expected by :func:`scipy.differentiate.hessian` and :func:`scipy.differentiate.jacobian`."""

    @wraps(func)
    def wrapped(x: np.ndarray) -> np.ndarray:
        inputs = np.moveaxis(x, 0, -1)
        raw_res = [func(inp) for inp in inputs.reshape(-1, x.shape[0])]
        out_shape = np.shape(raw_res[0])
        res = np.reshape(raw_res, x.shape[1:] + out_shape)  # pyright: ignore[reportCallIssue]
        if len(out_shape) != 0:
            return np.moveaxis(res, -1, 0)
        return res

    return wrapped


def vector_partial(
    func: Callable[[np.ndarray], np.ndarray], default: np.ndarray, dims: Iterable[int]
) -> Callable[[np.ndarray], np.ndarray]:
    """Partial function application for functions with a single vector valued argument.

    Args:
        func: Function to partially apply.
        default: Base vector that is partially applied. The values at the unbound dimensions are ignored.
        dims: The indices of the elements that are the inputs to the resulting function, the remaining
            values are taken from the default vector.

    Returns:
        A function that takes only the specified dimensions as input and fills in the rest from the default vector.
    """

    default = default.copy()

    @wraps(func)
    def wrapped(var: np.ndarray) -> np.ndarray:
        full_input = default.copy()
        for i, dim in enumerate(dims):
            full_input[dim] = var[i]
        return func(full_input)

    return wrapped


@dataclass(frozen=True)
class ModeMatchingAnalysis:
    """Analysis of a mode matching solution providing various sensitivity metrics."""

    solution: "ModeMatchingSolution"  #: The mode matching solution to analyze.

    @cached_property
    def hessian(self) -> np.ndarray:
        r"""The Hessian matrix :math:`\mathbf{H}` of the mode overlap function
        :math:`o(\mathbf{x})` around the optimum :math:`\mathbf{x}^*`. The individual elements
        :math:`h_{ij}` of the Hessian are given by:

        .. math::
            h_{ij} = \left. \frac{\partial^2 o(\mathbf{x})}{\partial x_i \partial x_j} \right|_{\mathbf{x} = \mathbf{x}^*}

        For problems with two degrees of freedom the Hessian is always negative definite, if there are
        more than two degrees of freedom it generally negative semi-definite with reduced rank or at least
        very bad conditioning.
        """

        mode_overlap = wrap_for_differentiate(
            self.solution.candidate.parametrized_overlap  # pyright: ignore[reportArgumentType]
        )

        # the default initial step is 0.5 which would lead to invalid lens position
        # 1e-2 ensures, that only physical configurations are evaluated
        tolerances = {"atol": 1e-6, "rtol": 1e-6}
        hess_res = hessian(mode_overlap, self.solution.positions, initial_step=1e-2, tolerances=tolerances)
        if np.any(hess_res.status != 0):
            warnings.warn(f"Hessian calculation did not converge: {hess_res.status}", stacklevel=2)

        return hess_res.ddf

    @cached_property
    def focus_and_waist_jacobian(self) -> np.ndarray:
        r"""The Jacobian :math:`\mathbf{J}` of the final beam focus and waist :math:`\mathbf{f}_{fw}(\mathbf{x})` with respect to
        the element positions :math:`\mathbf{x}` around the optimum :math:`\mathbf{x}^*`. The individual elements
        :math:`j_{ij}` of the Jacobian are given by:

        .. math::
            j_{ij} = \left.\frac{\partial f_{fw,i}(\mathbf{x})}{\partial x_j} \right|_{\mathbf{x} = \mathbf{x}^*}

        """
        # waist_and_focus = wrap_for_differentiate(make_focus_and_waist(self.solution))
        waist_and_focus = wrap_for_differentiate(self.solution.candidate.parametrized_focus_and_waist)
        jac_res = jacobian(waist_and_focus, self.solution.positions, initial_step=1e-2)
        if np.any(jac_res.status != 0):
            warnings.warn(f"Jacobian calculation did not converge: {jac_res.status}", stacklevel=2)
        return jac_res.df

    @cached_property
    def couplings(self) -> np.ndarray:
        r"""The coupling matrix :math:`\mathbf{R}` between the different degrees of freedom.

        The coupling :math:`r_{ij}` between degrees of freedom indexed :math:`i` and :math:`j` is
        the normalized cross-sensitivity :math:`s_{ij}` between the two degrees of freedom:

        .. math::
            r_{ij} = \frac{s_{ij}}{\sqrt{s_{ii} s_{jj}}}

        """
        normalizer = 1 / np.sqrt(np.diag(self.sensitivities))
        return self.sensitivities * np.outer(normalizer, normalizer)

    @cached_property
    def sensitivities(self) -> np.ndarray:
        r"""The sensitivity matrix :math:`\mathbf{S}` of the mode overlap around the optimum.

        It is proportional to the Hessian :math:`\mathbf{H}` with :math:`\mathbf{S} = -\mathbf{H}/2`.
        That way the positive loss in mode overlap :math:`\Delta o` for small perturbations
        :math:`\Delta \mathbf{x}` around the optimum can be expressed as:

        .. math::
            \Delta o \approx \mathbf{\Delta x}^T \mathbf{S} \mathbf{\Delta x}

        """

        return -self.hessian / 2

    @cached_property
    def min_coupling_pair(self) -> tuple[int, int]:
        """The indices :math:`(i, j)` of the pair of degrees of freedom with minimal absolute coupling :math:`r_{ij}`.
        The second index is always larger than the first.
        """

        indices = np.triu_indices(len(self.sensitivities), k=1)
        abs_couplings = np.abs(self.couplings[indices])
        best = np.argmin(abs_couplings)
        return (int(indices[0][best]), int(indices[1][best]))

    @cached_property
    def min_coupling(self) -> float:
        """The minimal absolute coupling between any pair of degrees of freedom."""
        return abs(float(self.couplings[self.min_coupling_pair]))

    @cached_property
    def min_cross_sens_pair(self) -> tuple[int, int]:
        """The indices :math:`(i, j)` of the pair of degrees of freedom with minimal absolute cross-sensitivity :math:`s_{ij}`.
        The second index is always larger than the first.
        """
        indices = np.triu_indices(len(self.sensitivities), k=1)
        abs_sensitivities = np.abs(self.sensitivities[indices])
        best = np.argmin(abs_sensitivities)
        return (int(indices[0][best]), int(indices[1][best]))

    @cached_property
    def min_cross_sens(self) -> float:
        """The minimal absolute cross-sensitivity between any pair of degrees of freedom."""
        return abs(float(self.sensitivities[self.min_cross_sens_pair]))

    @cached_property
    def min_cross_sens_direction(self) -> np.ndarray:
        """The direction of the least cross-sensitive pair of degrees of freedom.
        This is the smallest eigenvector of the 2x2 cross-sensitivity sub-matrix of the least cross-sensitive pair.
        """
        pair = self.min_cross_sens_pair
        eigs = np.linalg.eigh(self.sensitivities[np.ix_(pair, pair)])
        vec = eigs.eigenvectors[:, np.argmin(eigs.eigenvalues)]
        return vec if vec[0] >= 0 else -vec

    @cached_property
    def min_sensitivity_axis(self) -> int:
        """The index :math:`i` of the degree of freedom with minimal absolute sensitivity :math:`s_{ii}`."""

        diag_sensitivities = np.abs(np.diag(self.sensitivities))
        return int(np.argmin(diag_sensitivities))

    @cached_property
    def min_sensitivity(self) -> float:
        """The minimal absolute sensitivity among all degrees of freedom."""

        return float(self.sensitivities[self.min_sensitivity_axis, self.min_sensitivity_axis])

    @cached_property
    def max_sensitivity_axis(self) -> int:
        """The index :math:`i` of the degree of freedom with maximal absolute sensitivity :math:`s_{ii}`."""
        diag_sensitivities = np.abs(np.diag(self.sensitivities))
        return int(np.argmax(diag_sensitivities))

    @cached_property
    def max_sensitivity(self) -> float:
        """The maximal absolute sensitivity among all degrees of freedom."""

        return float(self.sensitivities[self.max_sensitivity_axis, self.max_sensitivity_axis])

    # the vectors spanning the sub space in which the mode overlap stays approximately constant
    # equivalent to the null space of the hessian assuming the minor eigenvalues are zero
    @cached_property
    def const_space(self) -> list[np.ndarray]:
        r"""The basis vectors spanning the constant overlap sub-space around the optimum.

        Note:
            This is simply determined as the corresponding eigenvectors two all but the two largest eigenvalues
            of the Hessian :math:`\mathbf{H}`. These eigenvalues are usually many orders of magnitude smaller than the
            two largest ones, but generally not zero.
        """

        eigs = np.linalg.eigh(self.hessian)
        return list(eigs.eigenvectors.T[np.argsort(eigs.eigenvalues)[2:]])

    @cached_property
    def grad_focus(self) -> np.ndarray:
        r"""The gradient of the final beam focus with respect to the element positions,
        equal to the first row of the Jacobian."""

        return self.focus_and_waist_jacobian[0]

    @cached_property
    def grad_waist(self) -> np.ndarray:
        r"""The gradient of the final beam waist with respect to the element positions,
        equal to the second row of the Jacobian."""
        return self.focus_and_waist_jacobian[1]

    def summary(self, sensitivity_unit: Config.SensitivityUnit | bool | None = None) -> dict:
        """Create a summary dictionary of the analysis results

        Args:
            sensitivity_unit: The unit to use for sensitivities in the summary.
                If `False`, the raw sensitivities without unit conversion are used.
                If `None` this defaults to :attr:`Config.sensitivity_unit <corset.config.Config.sensitivity_unit>`.

        Returns:
            A dictionary containing the summary data. The keys are:
            - "overlap": The mode overlap of the solution.
            - "num_elements": The number of free elements (i.e. elements used for mode matching) in the setup.
            - "elements": A list of the free elements (i.e. elements used for mode matching) in the setup.
            - "positions": The positions of the free elements in the setup.
            - "min_sensitivity_axis": The index of the degree of freedom with minimal sensitivity.
            - "min_sensitivity": The minimal sensitivity in the specified unit.
            - "max_sensitivity_axis": The index of the degree of freedom with maximal sensitivity.
            - "max_sensitivity": The maximal sensitivity in the specified unit.
            - "min_cross_sens_pair": The indices of the pair of degrees of freedom with minimal cross-sensitivity.
            - "min_cross_sens": The minimal cross-sensitivity in the specified unit.
            - "min_cross_sens_direction": The direction of the least cross-sensitive pair of degrees of freedom.
            - "min_coupling_pair": The indices of the pair of degrees of freedom with minimal coupling.
            - "min_coupling": The minimal coupling.
            - "sensitivities": The sensitivity matrix in the specified unit.
            - "couplings": The coupling matrix.
            - "const_space": The basis vectors spanning the constant overlap sub-space.
            - "grad_focus": The gradient of the final beam focus with respect to the element positions.
            - "grad_waist": The gradient of the final beam waist with respect to the element positions.
            - "sensitivity_unit": The sensitivity unit used.
            - "solution": The analyzed mode matching solution.

        """

        sensitivity_unit = cast(Config.SensitivityUnit, sensitivity_unit or Config.sensitivity_unit)
        factor = sensitivity_unit.value.factor
        sol = self.solution
        return {
            "overlap": sol.overlap,
            "num_elements": len(sol.positions),
            "elements": [sol.setup.elements[i][1] for i in sol.candidate.parametrized_setup.free_elements],
            "positions": sol.positions,
            "min_sensitivity_axis": self.min_sensitivity_axis,
            "min_sensitivity": self.min_sensitivity * factor,
            "max_sensitivity_axis": self.max_sensitivity_axis,
            "max_sensitivity": self.max_sensitivity * factor,
            "min_cross_sens_pair": self.min_cross_sens_pair,
            "min_cross_sens": self.min_cross_sens * factor,
            "min_cross_sens_direction": self.min_cross_sens_direction,
            "min_coupling_pair": self.min_coupling_pair,
            "min_coupling": self.min_coupling,
            "sensitivities": self.sensitivities * factor,
            "couplings": self.couplings,
            "const_space": self.const_space,
            "grad_focus": self.grad_focus,
            "grad_waist": self.grad_waist,
            "sensitivity_unit": sensitivity_unit,
            "solution": sol,
        }

    def summary_df(self, sensitivity_unit: Config.SensitivityUnit | None | bool = None) -> pd.DataFrame:
        """Create a summary DataFrame of the analysis results

        Args:
            sensitivity_unit: The unit to use for sensitivities in the summary. If `None` the default from :class:`Config` is used.
                If `False`, the raw sensitivities without unit conversion are used.

        Returns:
            A DataFrame containing the summary data with one row per value, see :meth:`summary` for details.
        """

        summary_data = self.summary(sensitivity_unit=sensitivity_unit)
        return pd.DataFrame([summary_data]).T



from importlib.metadata import PackageNotFoundError, version

from .config import Config  # noqa: F401
from .core import Beam, OpticalSetup, ThickLens, ThinLens  # noqa: F401
from .database import LensList  # noqa: F401
from .serialize import load_yaml, save_yaml  # noqa: F401
from .solver import (  # noqa: F401
    Aperture,
    ModeMatchingSolution,
    Passage,
    ShiftingRange,
    SolutionList,
    mode_match,
)

try:
    __version__ = version("corset")
except PackageNotFoundError:
    __version__ = "dev"



"""Module defining a LensList class for managing collections of lenses."""

from collections.abc import Iterable, Iterator
from dataclasses import dataclass, fields
from functools import cached_property
from itertools import chain
from pathlib import Path
from typing import cast, overload

import numpy as np
import pandas as pd

from corset.serialize import YamlSerializableMixin

from .core import Lens, ThickLens, ThinLens


@dataclass(frozen=True)
class LensList(YamlSerializableMixin):
    """A list of lenses with convenient access methods.

    Supports (array) indexing and other list-like operations, indices can also be
    strings or list of strings to access lenses by name.
    Implements :meth:`_repr_html_` to show a :class:`pandas.DataFrame` representation in IPython environments.
    """

    lenses: list[Lens]  #: Underlying list of lenses

    def __post_init__(self):
        self._by_name  # trigger validation  # noqa: B018

    def __iter__(self) -> Iterator[Lens]:
        return iter(self.lenses)

    def __len__(self) -> int:
        return len(self.lenses)

    @overload
    def __getitem__(self, index: int | str) -> Lens:
        pass

    @overload
    def __getitem__(self, index: slice | list[int] | list[str]) -> "LensList":
        pass

    def __getitem__(self, index: int | str | slice | list[int] | list[str]) -> Lens | "LensList":
        if isinstance(index, int):
            return self.lenses[index]
        elif isinstance(index, str):
            return self._by_name[index]
        elif isinstance(index, list) and len(index) > 0 and all(isinstance(i, str) for i in index):
            return LensList(lenses=[self._by_name[i] for i in index])  # pyright: ignore[reportArgumentType]
        elif isinstance(index, slice) or (isinstance(index, list) and all(isinstance(i, int) for i in index)):
            return LensList(lenses=np.array(self.lenses)[index].tolist())
        else:
            raise TypeError(f"Invalid index {index}")

    def concatenate(*lists: Iterable[Lens]) -> "LensList":
        """Concatenate multiple lens lists or lists of lenses.

        Args:
            lists: LensList or list of Lens instances to concatenate.

        Returns:
            LensList: The concatenated lens list.
        """
        return LensList(lenses=list(chain(*lists)))

    def __add__(self, other: Iterable[Lens]) -> "LensList":
        return LensList.concatenate(self, other)

    def __radd__(self, other: Iterable[Lens]) -> "LensList":
        return LensList.concatenate(other, self)

    @cached_property
    def _by_name(self) -> dict[str, Lens]:
        names = [lens.name for lens in self.lenses if lens.name]
        if len(names) != len(set(names)):
            raise ValueError("Lens names must be unique.")
        return {lens.name: lens for lens in self.lenses if lens.name}

    @cached_property
    def df(self) -> pd.DataFrame:
        """A :class:`pandas.DataFrame` representation of the lens list."""
        df = pd.DataFrame(
            [
                {
                    "name": lens.name,
                    "type": "thin" if not isinstance(lens, ThickLens) else "thick",
                    "focal_length": lens.focal_length,
                    "left_margin": lens.left_margin,
                    "right_margin": lens.right_margin,
                    "in_roc": vars(lens).get("in_roc"),
                    "out_roc": vars(lens).get("out_roc"),
                    "thickness": vars(lens).get("thickness"),
                    "refractive_index": vars(lens).get("refractive_index"),
                    "lens": lens,
                }
                for lens in self.lenses
            ]
        )
        if all(df["type"] == "thin"):
            df = df.drop(columns=["in_roc", "out_roc", "thickness", "refractive_index"])
        return df

    def _repr_html_(self) -> str:
        return self.df.to_html(notebook=True)

    def save_csv(self, path: str | Path) -> None:
        """Save the lens list to a CSV file.

        Args:
            path: Path to the CSV file. This is passed directly to :func:`pandas.DataFrame.to_csv` so it also
                accepts other types supported by that function like file-like objects.
        """
        self.df.drop(columns=["lens"]).to_csv(path, index=False)

    @classmethod
    def load_csv(cls, path: str | Path) -> "LensList":
        """Load a lens list from a CSV file.

        Args:
            path: Path to the CSV file. This is passed directly to :func:`pandas.read_csv` so it also
                accepts other types supported by that function like URL strings or file-like objects.

        Returns:
            LensList: The loaded lens list.

        Raises:
            ValueError: If the CSV file contains invalid data.
            KeyError: If required fields are missing in the CSV file.
        """
        df = pd.read_csv(path)
        lenses: list[Lens] = []
        for i, row in df.iterrows():
            line = cast(int, i) + 2
            lens_cls = {"thin": ThinLens, "thick": ThickLens}.get(row["type"])
            if lens_cls is None:
                raise ValueError(f"Unknown lens type: {row['type']}")
            try:
                kwargs = {f.name: row[f.name] for f in fields(lens_cls)}
                kwargs["name"] = kwargs["name"] if pd.notna(kwargs["name"]) else None
                lens: Lens = lens_cls(**kwargs)  # type: ignore[arg-type]
                lenses.append(lens)
            except ValueError as e:
                raise ValueError(f"Error creating lens from line {line}: {e}") from e
            except KeyError as e:
                raise ValueError(f"Missing required field '{e.args[0]}' for lens from line {line}") from e
            if lens_cls is ThickLens and not np.isclose(lens.focal_length, row.get("focal_length", lens.focal_length)):
                raise ValueError(f"If specified, thick lens focal length must match computed value (line {line})")
        return cls(lenses=lenses)

    @classmethod
    def available(cls) -> list[str]:
        """List available built-in lens lists.

        Returns:
            list[str]: Names of available lens lists.
        """
        raise NotImplementedError("Built-in lens library not implemented yet.")

    @classmethod
    def load(cls, name: str) -> "LensList":
        """Load a lens list from the built-in library.

        Args:
            name: Name of the lens list to load.

        Returns:
            LensList: The loaded lens list.

        Raises:
            ValueError: If the lens list name is not found in the library.
        """
        raise NotImplementedError("Built-in lens library not implemented yet.")



# -- Path setup --------------------------------------------------------------
import os
import sys

sys.path.insert(0, os.path.abspath("../../src"))

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Beam Corset"
copyright = "2025, Lorenz Kies"
author = "Lorenz Kies"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx.ext.intersphinx",
    "nbsphinx",
    "sphinx_copybutton",
]

intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "IPython": ("https://ipython.readthedocs.io/en/stable/", None),
}

templates_path = ["_templates"]
exclude_patterns = []

autodoc_member_order = "groupwise"
autodoc_typehints = "both"
typehints_use_signature_return = True
typehints_defaults = "comma"

nbsphinx_execute = "always"


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
# html_static_path = ["_static"]

html_theme_options = {
    "logo": {
        "text": "Beam Corset",
        "image_light": "../../misc/logo/logo_light.svg",
        "image_dark": "../../misc/logo/logo_dark.svg",
    },
}

html_favicon = "../../misc/logo/favicon.svg"


### Preprocess notebooks for RST compatibility
# TODO move this to a separate script and check timestamps to only preprocess when necessary
import json
from pathlib import Path


def process_notebook(source: Path, output: Path):
    notebook = json.loads(source.read_text(encoding="utf-8"))

    for cell in notebook.get("cells", []):
        metadata = cell.get("metadata", {})
        if cell.get("cell_type") == "markdown" and "rst" in metadata.get("tags", []):
            cell["cell_type"] = "raw"
            cell["metadata"] = metadata | {"raw_mimetype": "text/restructuredtext"}
            cell["source"] = [line.replace("\\", "\\\\") for line in cell["source"]]

    output.parent.mkdir(parents=True, exist_ok=True)
    output.write_text(json.dumps(notebook, indent=1), encoding="utf-8")


def preprocess_notebooks(source_dir: Path, output_dir: Path):
    for source in source_dir.glob("*.ipynb"):
        output = output_dir / source.name
        process_notebook(source, output)


this_file_dir = Path(__file__).parent

preprocess_notebooks(this_file_dir / "../notebooks", this_file_dir / "gen")



"""Plotting functions for optical setups and mode matching solutions and analyses."""

from dataclasses import dataclass
from io import BytesIO
from itertools import product
from typing import TYPE_CHECKING, Any, cast

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.collections import (
    FillBetweenPolyCollection,
    LineCollection,
    PathCollection,
)
from matplotlib.colorbar import Colorbar
from matplotlib.figure import Figure
from matplotlib.lines import Line2D
from matplotlib.patches import Ellipse, Rectangle
from matplotlib.text import Annotation
from matplotlib.ticker import FuncFormatter
from scipy import stats

from .config import Config

# prevent circular import for type annotation in function signature
if TYPE_CHECKING:
    from .core import OpticalSetup
    from .solver import ModeMatchingSolution

RELATIVE_MARGIN = 0.1  #: Relative margin size for plotting optical setups

milli_formatter = FuncFormatter(lambda x, _: f"{x*1e3:.0f}")  #: Formatter for millimeter axes
micro_formatter = FuncFormatter(lambda x, _: f"{x*1e6:.0f}")  #: Formatter for micrometer axes


def fig_to_png(fig: Figure) -> bytes:
    """Convert a Matplotlib figure to a PNG as bytes.

    Args:
        fig: The figure to convert.

    Returns:
        The PNG representation of the figure as bytes.
    """
    buf = BytesIO()
    fig.savefig(buf, format="png", bbox_inches="tight")
    plt.close(fig)
    return buf.getvalue()


@dataclass(frozen=True)
class OpticalSetupPlot:
    """Plot of an optical setup with references to the plot elements."""

    ax: Axes  #: Axes containing the plot
    zs: np.ndarray  #: Z-coordinates of the beam profile
    rs: np.ndarray  #: Radii of the beam profile
    beam: FillBetweenPolyCollection  #: Fill between collection for the beam
    beam_ci: list[Line2D]  #: List of lines representing beam confidence intervals
    lenses: list[tuple[LineCollection, Annotation, Rectangle | None]]  #: List of lens plot elements
    handles: list[tuple[str, Any]]  #: List of plot handles by their legend labels
    r_max: float  #: Maximum beam radius in the plot


@dataclass(frozen=True)
class ModeMatchingPlot:
    """Plot of a mode matching solution with references to the plot elements."""

    ax: Axes  #: Axes containing the plot
    setup: OpticalSetupPlot  #: Plot of the actual optical setup
    desired_beam: OpticalSetupPlot  #: Plot of the desired beam profile
    ranges: list[Rectangle]  #: List of ranges in the plot
    apertures: list[Line2D]  #: List of aperture lines in the plot
    passages: list[Rectangle]  #: List of passage rectangles in the plot


def get_handles(ax: Axes) -> list[tuple[str, Any]]:
    """Get a list of plot handles by their legend labels.

    Args:
        ax: The axes to get the handles from.

    Returns:
        A list of tuples containing legend labels and their corresponding plot handles, empty if no legend exists.
    """

    legend = ax.get_legend()
    if legend is None:
        return []
    return [(text.get_text(), handle) for handle, text in zip(legend.legend_handles, legend.texts, strict=True)]


# TODO refactor this function to reduce complexity
def plot_setup(  # noqa: C901
    self: "OpticalSetup",
    *,
    points: int | np.ndarray | None = None,
    limits: tuple[float, float] | None = None,
    beam_kwargs: dict | None = None,
    confidence_interval: float | bool | None = None,
    rayleigh_range_cap: float | None = None,
    free_lenses: list[int] = [],  # noqa: B006
    ax: Axes | None = None,
    show_legend: bool | None = None,
    legend_loc: str | None = None,
    # TODO add back other configuration options?
) -> OpticalSetupPlot:
    """Plot the the beam profile and optical elements of the setup.

    Args:
        self: The optical setup instance.
        points: Number of points or specific z-coordinates to evaluate the beam profile.
            If `None`, this defaults to :attr:`Config.PlotSetup.beam_points <corset.config.Config.PlotSetup.beam_points>`.
        limits: Z-coordinate limits for the plot.
            If `None`, limits are determined from the beam and elements.
        beam_kwargs: Additional keyword arguments passed to the beam plot.
            If `None` this defaults to :attr:`Config.PlotSetup.beam_kwargs <corset.config.Config.PlotSetup.beam_kwargs>`.
        confidence_interval: Confidence interval probability for beam uncertainty visualization.
            If `False`, no uncertainty is plotted.
            If `None` this defaults to :attr:`Config.PlotSetup.confidence_interval <corset.config.Config.PlotSetup.confidence_interval>`.
        rayleigh_range_cap: Maximum Rayleigh range to consider when determining plot limits.
            If `None` this defaults to :attr:`Config.PlotSetup.rayleigh_range_cap <corset.config.Config.PlotSetup.rayleigh_range_cap>`.
        free_lenses: Indices of lenses to treat as free elements in the plot.
        ax: The axes to plot on. If `None`, the current axes are used.
        show_legend: Whether to show a legend for the plot.
            If `None`, this defaults to :attr:`Config.PlotSetup.show_legend <corset.config.Config.PlotSetup.show_legend>`.
        legend_loc: Location of the legend in the plot.
            If `None`, this defaults to :attr:`Config.PlotSetup.legend_loc <corset.config.Config.PlotSetup.legend_loc>`.

    Returns:
        An :class:`OpticalSetupPlot` containing references to the plot elements.
    """

    ax = ax or plt.gca()
    show_legend = Config.get(show_legend, Config.PlotSetup.show_legend)
    legend_loc = Config.get(legend_loc, Config.PlotSetup.legend_loc)
    beam_kwargs = Config.get(beam_kwargs, Config.PlotSetup.beam_kwargs)
    if confidence_interval is True:
        raise ValueError("confidence_interval cannot be True, must be a float between 0 and 1 or False")
    confidence_interval = Config.get(confidence_interval, Config.PlotSetup.confidence_interval)
    rayleigh_range_cap = Config.get(rayleigh_range_cap, Config.PlotSetup.rayleigh_range_cap)

    lens_positions = [pos for pos, _ in self.elements]

    if isinstance(points, np.ndarray):
        zs = points
    else:
        if not limits:
            cap = rayleigh_range_cap
            all_bounds = [
                self.beams[0].focus - min(cap, self.beams[0].rayleigh_range),
                self.beams[-1].focus + min(cap, self.beams[-1].rayleigh_range),
                *(),
                *lens_positions,
            ]
            if self.elements:
                all_bounds.append(self.elements[0][0] - min(cap, self.beams[0].rayleigh_range))
                all_bounds.append(self.elements[-1][0] + min(cap, self.beams[-1].rayleigh_range))

            limits = (min(all_bounds), max(all_bounds))

        num_points = points if isinstance(points, int) else Config.PlotSetup.beam_points
        zs = np.linspace(limits[0], limits[1], num_points)

    handles = get_handles(ax)

    rs = cast(np.ndarray, self.radius(zs))
    fill_between = ax.fill_between(zs, -rs, rs, **{"zorder": 100, **beam_kwargs})
    beam_label = beam_kwargs.get("label", "Beam")
    handles.append((beam_label, fill_between))
    r_max = np.max(rs)

    beam_deviation = []
    if confidence_interval is not False and self.beams[0].gauss_cov is not None:
        rs_ci = self.radius_dev(zs) * stats.norm.interval(confidence_interval)[1]
        ci_color = beam_kwargs.get("color")
        if ci_color is None or ci_color == "none":
            ci_color = beam_kwargs.get("edgecolor")
        for r0, ci in product([-rs, rs], [-rs_ci, rs_ci]):
            line = ax.plot(zs, r0 + ci, ls="--", color=ci_color, alpha=beam_kwargs.get("alpha"), zorder=105)[0]
            beam_deviation.append(line)
        r_max = np.max(rs + rs_ci)
        handles.append((f"{round(confidence_interval*100)}% CI ({beam_label})", beam_deviation[0]))
    # TODO make beam plot function?

    # TODO factor out into plot lens function?
    lenses = []
    for i, (pos, lens) in enumerate(self.elements):
        color = "C2"
        zorder = 200

        line = ax.vlines(
            pos,
            -r_max * (1 + RELATIVE_MARGIN),
            r_max * (1 + RELATIVE_MARGIN),
            color=color,
            zorder=zorder,
        )

        label_text = lens.name if lens.name is not None else f"f={round(lens.focal_length*1e3)}mm"
        if i in free_lenses:
            label_text = f"$L_{i}$: {label_text} @{round(pos*1e3)}mm"
        label = ax.text(
            pos,
            -r_max * (1 + RELATIVE_MARGIN),
            label_text,
            va="center",
            ha="left",
            rotation="vertical",
            rotation_mode="anchor",
            bbox={"fc": plt.rcParams["axes.facecolor"], "ec": "none", "alpha": 0.5},
            zorder=zorder,
        )

        rect = None
        if lens.left_margin or lens.right_margin:
            rect = Rectangle(
                (pos - lens.left_margin, -r_max * (1 + RELATIVE_MARGIN)),
                lens.left_margin + lens.right_margin,
                2 * r_max * (1 + RELATIVE_MARGIN),
                fc="none",
                ec=color,
                ls="--",
                zorder=zorder,
            )
            ax.add_patch(rect)

        lenses.append((line, label, rect))

    if lenses and not any(label == "Lens" for label, _ in handles):
        # fake handle with vertical line marker
        handles.append(("Lens", Line2D([], [], color=color, label="Lens", marker="|", linestyle="None", markersize=10)))
    rectangles = [r for _, _, r in lenses if r is not None]
    if rectangles and not any(label == "Lens Margin" for label, _ in handles):
        handles.append(("Lens Margin", rectangles[0]))

    ax.set_ylim(
        min(-r_max * (1 + 3 * RELATIVE_MARGIN), ax.get_ylim()[0]),
        max(r_max * (1 + 3 * RELATIVE_MARGIN), ax.get_ylim()[1]),
    )

    ax.set_xlabel("z in mm")
    ax.xaxis.set_major_formatter(milli_formatter)

    ax.set_ylabel(r"w(z) in $\mathrm{\mu m}$")
    ax.yaxis.set_major_formatter(micro_formatter)

    if show_legend:
        ax.legend([lbl for _, lbl in handles], [han for han, _ in handles], loc=legend_loc).set_zorder(1000)

    return OpticalSetupPlot(
        ax=ax,
        zs=zs,
        rs=rs,
        beam=fill_between,
        beam_ci=beam_deviation,
        lenses=lenses,
        handles=handles,
        r_max=r_max,
    )


def plot_mode_match_solution_setup(
    self: "ModeMatchingSolution",
    *,
    ax: Axes | None = None,
    show_legend: bool | None = None,
    legend_loc: str | None = None,
) -> ModeMatchingPlot:
    """Plot the mode matching solution setup including the desired beam and constraints.

    Args:
        self: The mode matching solution instance.
        ax: The axes to plot on. If `None`, the current axes are used.
        show_legend: Whether to show a legend for the plot.
            If `None`, this defaults to :attr:`Config.PlotSetup.show_legend <corset.config.Config.PlotSetup.show_legend>`.
        legend_loc: Location of the legend in the plot.
            If `None`, this defaults to :attr:`Config.PlotSetup.legend_loc <corset.config.Config.PlotSetup.legend_loc>`.
    Returns:
        An :class:`ModeMatchingPlot` containing references to the plot elements.
    """
    from .core import OpticalSetup
    from .solver import Aperture, Passage

    ax = ax or plt.gca()
    show_legend = Config.get(show_legend, Config.PlotSetup.show_legend)
    legend_loc = Config.get(legend_loc, Config.PlotSetup.legend_loc)

    problem = self.candidate.problem

    setup_plot = plot_setup(
        self.setup, ax=ax, free_lenses=self.candidate.parametrized_setup.free_elements, show_legend=show_legend
    )
    desired_plot = plot_setup(
        OpticalSetup(problem.desired_beam, []),
        ax=ax,
        beam_kwargs={"color": "C1", "label": "Desired Beam"},
        show_legend=show_legend,
    )

    handles = get_handles(ax)
    if show_legend:
        desired = handles.pop(-1)
        assert desired[0] == "Desired Beam"  # noqa: S101
        handles.insert(1, desired)  # hack to show desired beam 2nd in legend

    ranges = []
    for range_ in problem.ranges:
        # TODO color configuratbility?
        rectangle = Rectangle(
            (range_.left, -setup_plot.r_max * (1 + 2 * RELATIVE_MARGIN)),
            range_.right - range_.left,
            2 * setup_plot.r_max * (1 + 2 * RELATIVE_MARGIN),
            fc="none",
            ec="lightgrey",
            ls="--",
            zorder=50,  # TODO
        )
        ranges.append(ax.add_patch(rectangle))
    if ranges:
        handles.append(("Shifting Range", ranges[0]))

    apertures = []
    passages = []
    for con in problem.constraints:
        if isinstance(con, Aperture):
            apertures.extend(
                ax.plot([con.position, con.position], [-con.radius, con.radius], marker="o", color="C4", zorder=80)
            )
        elif isinstance(con, Passage):
            rect = Rectangle(
                (con.left, -con.radius),
                con.right - con.left,
                2 * con.radius,
                fc="C4",
                ec="none",
            )
            passages.append(ax.add_patch(rect))

    if apertures:
        # TODO add dots to vertical line?
        handles.append(  # fake handle with line marker
            ("Aperture", Line2D([], [], color="C4", label="Aperture", marker="|", linestyle="None", markersize=10))
        )
    if passages:
        handles.append(("Passage", passages[0]))

    if show_legend:
        ax.legend([lbl for _, lbl in handles], [han for han, _ in handles], loc=legend_loc).set_zorder(1000)

    ax.set_title(f"Optical Setup ({self.overlap*100:.2f}% mode overlap)")

    return ModeMatchingPlot(
        ax=ax,
        setup=setup_plot,
        desired_beam=desired_plot,
        ranges=ranges,
        apertures=apertures,
        passages=passages,
    )


@dataclass(frozen=True)
class ReachabilityPlot:
    """Plot of a reachability analysis with references to the plot elements."""

    ax: Axes  #: Axes containing the plot
    center: PathCollection  #: Point representing the optimal focus and waist
    center_ci: Ellipse | None  #: Uncertainty of the arriving beam around the optimal point
    lines: list[list[Line2D]]  #: Lines representing parameter variations
    contour: Any  #: Contour plot of the mode overlap
    colorbar: Colorbar | None  #: Colorbar for the contour plot


@dataclass(frozen=True)
class SensitivityPlot:
    """Plot of a sensitivity analysis with references to the plot elements."""

    ax: Axes  #: Axes containing the plot
    displacements_ci: Ellipse | None  #: Confidence ellipse in terms of required displacements
    contours: list[Any]  #: Contour plots of the sensitivity
    colorbar: Colorbar | None  #: Colorbar for the contour plots
    handles: list  #: Handles for the plot elements


def plot_ellipse(
    ax: Axes, mean: tuple[float, float], var: np.ndarray, confidence_interval: float, **ellipse_kwargs
) -> Ellipse:
    eigenvalues, eigenvectors = np.linalg.eigh(var)
    angle = np.degrees(np.arctan2(*eigenvectors[:, 0][::-1]))
    diameters = 2 * np.sqrt(eigenvalues) * stats.chi2.ppf(confidence_interval, df=2) ** 0.5
    base_kwargs = {"fill": False, "ls": "--", "zorder": 100, "color": plt.rcParams["text.color"]}
    patch = Ellipse(mean, *diameters, angle=angle, **(base_kwargs | ellipse_kwargs))
    ax.add_patch(patch)
    return patch


def plot_reachability(
    self: "ModeMatchingSolution",
    *,
    displacement: float | list[float] | None = None,
    num_samples: int | list[int] | None = None,
    line_step: int | list[int] | None = None,
    dimensions: list[int] | None = None,
    focus_range: tuple[float, float] | None = None,
    waist_range: tuple[float, float] | None = None,
    confidence_interval: float | None = None,
    grid_resolution: int | None = None,
    ax: Axes | None = None,
) -> ReachabilityPlot:
    """Plot a reachability analysis of the mode matching solution.

    Args:
        self: The mode matching solution instance.
        displacement: Maximum displacement from the optimal position(s) in meters.
            If `None`, this defaults to :attr:`Config.PlotReachability.displacement <corset.config.Config.PlotReachability.displacement>`.
        num_samples: Number of samples to take along each dimension.
            If `None`, this defaults to :attr:`Config.PlotReachability.num_samples <corset.config.Config.PlotReachability.num_samples>`.
        line_step: Step size to skip certain lines for increased smoothness while retaining clarity.
            If `None`, this defaults to :attr:`Config.PlotReachability.line_step <corset.config.Config.PlotReachability.line_step>`.
        dimensions: Indices of the dimensions to analyze. If `None`, all dimensions are used.
        focus_range: Range of focus values to display on the x-axis. If `None`, determined automatically.
        waist_range: Range of waist values to display on the y-axis. If `None`, determined automatically.
        confidence_interval: Confidence ellipse probability for focus position and waist radius.
            If `None`, this defaults to :attr:`Config.PlotReachability.confidence_interval <corset.config.Config.PlotReachability.confidence_interval>`.
        grid_resolution: Resolution of the grid for the background mode overlap contour plot.
            If `None`, this defaults to :attr:`Config.Overlap.grid_resolution <corset.config.Config.Overlap.grid_resolution>`.
        ax: The axes to plot on. If `None`, the current axes are used.

    Returns:
        A :class:`ReachabilityPlot` containing references to the plot elements.
    """

    from .analysis import vector_partial
    from .solver import mode_overlap

    ax = ax or plt.gca()
    displacement = Config.get(displacement, Config.PlotReachability.displacement)
    num_samples = Config.get(num_samples, Config.PlotReachability.num_samples)
    line_step = Config.get(line_step, Config.PlotReachability.line_step)
    confidence_interval = Config.get(confidence_interval, Config.PlotReachability.confidence_interval)
    grid_resolution = Config.get(grid_resolution, Config.Overlap.grid_resolution)

    if dimensions is None:
        dimensions = list(range(len(self.positions)))

    num_dof = len(dimensions)
    displacement = cast(list, [displacement] * num_dof if np.isscalar(displacement) else displacement)
    num_samples = cast(list, [num_samples] * num_dof if np.isscalar(num_samples) else num_samples)
    line_step = cast(list, [line_step] * num_dof if np.isscalar(line_step) else line_step)

    linspaces = [
        np.linspace(-d, d, num=n) + offset
        for d, n, offset in zip(
            displacement, num_samples, self.positions, strict=True  # pyright: ignore[reportArgumentType]
        )
    ]

    if focus_range is not None:
        ax.set_xlim(focus_range)

    if waist_range is not None:
        ax.set_ylim(waist_range)

    focus_and_waist = np.vectorize(
        vector_partial(self.candidate.parametrized_focus_and_waist, self.positions, dimensions), signature="(n)->(2)"
    )

    grids = np.moveaxis(np.meshgrid(*linspaces, indexing="ij"), 0, -1)  # pyright: ignore[reportArgumentType]
    focuses, waists = np.moveaxis(focus_and_waist(grids), -1, 0)
    delta_focuses = focuses - self.candidate.problem.desired_beam.focus

    contrast_color = plt.rcParams["text.color"]
    center = ax.scatter(0, self.candidate.problem.desired_beam.waist, color=contrast_color, marker="o", zorder=100)
    center_ci = None
    if confidence_interval is not False and self.setup.beams[-1].gauss_cov is not None:
        center_ci = plot_ellipse(
            ax,
            (0, self.candidate.problem.desired_beam.waist),
            self.setup.beams[-1].gauss_cov,
            confidence_interval,
        )

    def select_lines(arr: np.ndarray, dim: int, steps: list[int]) -> np.ndarray:
        steps = [step if i != dim else 1 for i, step in enumerate(steps)]
        indices = tuple(slice(None, None, step) for step in steps)
        return np.moveaxis(arr[indices], dim, -1).reshape(-1, arr.shape[dim])

    lines: list[list[Line2D]] = []
    for i, (dim, disp) in enumerate(zip(dimensions, displacement, strict=True)):
        focuses_flat = select_lines(delta_focuses, i, line_step)  # pyright: ignore[reportArgumentType]
        waists_flat = select_lines(waists, i, line_step)  # pyright: ignore[reportArgumentType]
        lines.append([])
        for focus, waist in zip(focuses_flat, waists_flat, strict=True):
            line = ax.plot(focus, waist, color=f"C{dim}", label=rf"$\Delta x_{i}$ ($\pm{disp*1e3:.1f}$ mm)")[0]
            lines[-1].append(line)
        ax.annotate(
            "",
            xy=(focuses_flat[0][1], waists_flat[0][1]),
            xytext=(focuses_flat[0][0], waists_flat[0][0]),
            arrowprops={"color": f"C{i}", "headwidth": 10, "width": 2},
        )

    ax.legend(handles=[line[0] for line in lines], loc="lower left").set_zorder(1000)

    focuses_grid, waists_grid = np.meshgrid(
        np.linspace(*ax.get_xlim(), grid_resolution), np.linspace(*ax.get_ylim(), grid_resolution)
    )
    overlap = mode_overlap(
        focuses_grid,  # pyright: ignore[reportArgumentType]
        self.candidate.problem.desired_beam.waist,
        waists_grid,  # pyright: ignore[reportArgumentType]
        self.candidate.problem.setup.initial_beam.wavelength,
    )

    levels = Config.Overlap.levels
    colors = Config.Overlap.colors()
    res = ax.contourf(focuses_grid, waists_grid, overlap * 100, levels=levels, colors=colors, alpha=0.5)
    cb = ax.figure.colorbar(res, label="Mode overlap (%)")

    jacobian = self.analysis.focus_and_waist_jacobian

    ax.set_xlabel(rf"$\Delta z_0$ in mm ($\nabla z_0$=[{' '.join(f'{x:.3f}' for x in jacobian[0])}])")
    ax.xaxis.set_major_formatter(milli_formatter)

    ax.set_ylabel(rf"$w_0$ in $\mathrm{{\mu m}}$ ($\nabla w_0$=[{' '.join(f'{x:.3f}' for x in jacobian[1]*1e3)}]e-3)")
    ax.yaxis.set_major_formatter(micro_formatter)

    ax.set_title("Reachability Analysis")

    return ReachabilityPlot(ax=ax, center=center, center_ci=center_ci, lines=lines, contour=res, colorbar=cb)


def plot_sensitivity(
    self: "ModeMatchingSolution",
    *,
    dimensions: tuple[int, int] | tuple[int, int, int] | None = None,
    worst_overlap: float | None = None,
    x_displacement: float | None = None,
    y_displacement: float | None = None,
    z_displacement: float | None = None,
    num_samples_z: int | None = None,
    confidence_interval: float | None = None,
    force_contour_lines: bool | None = None,
    grid_resolution: int | None = None,
    ax: Axes | None = None,
) -> SensitivityPlot:
    """Plot a sensitivity analysis of the mode matching solution.

    Args:
        self: The mode matching solution instance.
        dimensions: Indices of the dimensions to analyze. If `None`, the two least
            coupled dimensions are used as the x and y dimensions, and the remaining most sensitive
            dimension is used as the auxiliary z dimension if applicable.
        worst_overlap: Worst-case mode overlap contour line that should still be fully visible in the
            plot used for automatic range determination.
            If `None`, this defaults to :attr:`Config.PlotSensitivity.worst_overlap <corset.config.Config.PlotSensitivity.worst_overlap>`.
        x_displacement: Displacement magnitude for x-axis values. If `None`, determined automatically from worst_overlap.
        y_displacement: Displacement magnitude for y-axis values. If `None`, determined automatically from worst_overlap.
        z_displacement: Displacement magnitude for z-axis values. If `None`, determined automatically from worst_overlap.
        num_samples_z: Number of z-slices to plot if 3 dimensions are used.
            If `None`, this defaults to :attr:`Config.PlotSensitivity.num_samples_z <corset.config.Config.PlotSensitivity.num_samples_z>`.
        confidence_interval: Confidence ellipse probability in terms of the required x and y displacements.
            If `None`, this defaults to :attr:`Config.PlotSensitivity.confidence_interval <corset.config.Config.PlotSensitivity.confidence_interval>`.
        force_contour_lines: If `True`, always use contour lines instead of filled contours for plots with two
            degrees of freedom. If `None`, this defaults to :attr:`Config.PlotSensitivity.force_contour_lines <corset.config.Config.PlotSensitivity.force_contour_lines>`.
        grid_resolution: Resolution of the grid for the contour plots.
            If `None`, this defaults to :attr:`Config.Overlap.grid_resolution <corset.config.Config.Overlap.grid_resolution>`.
        ax: The axes to plot on. If `None`, the current axes are used.

    Returns:
        A :class:`SensitivityPlot` containing references to the plot elements.
    """

    from .analysis import vector_partial

    worst_overlap = Config.get(worst_overlap, Config.PlotSensitivity.worst_overlap)
    num_samples_z = Config.get(num_samples_z, Config.PlotSensitivity.num_samples_z)
    confidence_interval = Config.get(confidence_interval, Config.PlotSensitivity.confidence_interval)
    force_contour_lines = Config.get(force_contour_lines, Config.PlotSensitivity.force_contour_lines)
    grid_resolution = Config.get(grid_resolution, Config.Overlap.grid_resolution)
    ax = ax or plt.gca()

    if dimensions is None:
        dimensions = self.analysis.min_coupling_pair
        if len(self.positions) > 2:
            all_dims = set(range(len(self.positions)))
            remaining_dims = list(all_dims - set(dimensions))
            most_sensitive = max(remaining_dims, key=lambda dim: self.analysis.sensitivities[dim, dim])
            dimensions = (*dimensions, most_sensitive)

    mode_overlap = np.vectorize(
        vector_partial(
            self.candidate.parametrized_overlap, self.positions, dimensions  # pyright: ignore[reportArgumentType]
        ),
        signature="(n)->()",
    )

    # the hessian will always be badly conditioned for ndim > 2, so only use the subspace we care about
    # which should be well conditioned
    A_xy_inv = np.linalg.inv(-self.analysis.hessian[np.ix_(dimensions[:2], dimensions[:2])] / 2)  # noqa: N806
    # use Config.get so the type checker knows the displacements are not None after this point
    x_displacement = Config.get(x_displacement, np.sqrt(A_xy_inv[0, 0] * (1 - worst_overlap)) * 1.1)
    y_displacement = Config.get(y_displacement, np.sqrt(A_xy_inv[1, 1] * (1 - worst_overlap)) * 1.1)

    if z_displacement is None and len(dimensions) > 2:
        A_xyz = -self.analysis.hessian[np.ix_(dimensions, dimensions)]  # noqa: N806
        eigs = np.linalg.eigh(A_xyz)
        v_min = eigs.eigenvectors[:, np.argmin(eigs.eigenvalues)]
        z_displacement = min(abs(v_min[2] / v_min[0] * x_displacement), abs(v_min[2] / v_min[1] * y_displacement)) / 1.1

    base = np.array([self.positions[dim] for dim in dimensions])
    xs = np.linspace(-x_displacement, x_displacement, grid_resolution)
    ys = np.linspace(-y_displacement, y_displacement, grid_resolution)
    xsg, ysg = np.meshgrid(xs, ys)
    cmap = plt.get_cmap("coolwarm") if Config.mpl_is_dark() else plt.get_cmap("berlin")

    contours = []
    colorbar = None
    handles: list = []

    if len(dimensions) == 2:
        overlaps = mode_overlap(np.stack([xsg, ysg], axis=-1) + base)
        if force_contour_lines:
            cont = ax.contour(xsg, ysg, overlaps * 100, levels=Config.Overlap.levels, colors=cmap(0.5))
            ax.clabel(cont, fmt="%1.1f%%")
            contours.append(cont)
        else:
            cont = ax.contourf(xsg, ysg, overlaps * 100, levels=Config.Overlap.levels, colors=Config.Overlap.colors())
            colorbar = ax.figure.colorbar(cont, label="Mode overlap (%)")
            contours.append(cont)
    else:
        z_displacement = cast(float, z_displacement)  # make type checker happy
        z_colors = cmap(np.linspace(0, 1, num_samples_z))
        zs = np.linspace(-z_displacement, z_displacement, num_samples_z)
        for i, (color, z) in enumerate(zip(z_colors, zs, strict=True)):
            positions = np.stack([xsg, ysg, z * np.ones_like(xsg)], axis=-1) + base
            overlaps = mode_overlap(positions)
            zorder = 100 if i == num_samples_z // 2 else 10
            cont = ax.contour(
                xsg, ysg, overlaps * 100, levels=Config.Overlap.levels, colors=[color], alpha=0.7, zorder=zorder
            )
            contours.append(cont)
            if i == num_samples_z // 2:
                ax.clabel(cont, fmt="%1.1f%%", colors=[color])
        handles = [
            Line2D([], [], color=color, label=rf"$\Delta x_{{{dimensions[2]}}} = {value*1e3:.1f}$ mm")
            for color, value in zip(z_colors, zs, strict=True)
        ]
        ax.legend(handles=handles, loc="lower left").set_zorder(1000)

    displacements_ci = None
    if confidence_interval is not False and self.setup.beams[-1].gauss_cov is not None:
        jac_inv = np.linalg.inv(self.analysis.focus_and_waist_jacobian[:, dimensions[:2]])
        cov_xy = jac_inv @ self.setup.beams[-1].gauss_cov @ jac_inv.T
        displacements_ci = plot_ellipse(ax, (0, 0), cov_xy, confidence_interval)

    unit = Config.sensitivity_unit
    dims = dimensions
    sens_x = self.analysis.sensitivities[dims[0], dims[0]] * unit.value.factor
    sens_y = self.analysis.sensitivities[dims[1], dims[1]] * unit.value.factor
    ax.set_xlabel(rf"$\Delta x_{{{dims[0]}}}$ in mm ($s_{{{str(dims[0])*2}}}={sens_x:.2f}{unit.value.tex}$)")
    ax.set_ylabel(rf"$\Delta x_{{{dims[1]}}}$ in mm ($s_{{{str(dims[1])*2}}}={sens_y:.2f}{unit.value.tex}$)")
    ax.xaxis.set_major_formatter(milli_formatter)
    ax.yaxis.set_major_formatter(milli_formatter)

    ax.set_title(rf"Sensitivity Analysis ($r_{{{dims[0]}{dims[1]}}}={self.analysis.min_coupling*100:.2f}\%$)")

    return SensitivityPlot(
        ax=ax, displacements_ci=displacements_ci, contours=contours, colorbar=colorbar, handles=handles
    )


def plot_mode_match_solution_all(
    self: "ModeMatchingSolution",
    *,
    figsize: tuple[float, float] | None = None,
    width_ratios: tuple[float, float, float] | None = None,
    tight_layout: bool | None = None,
    setup_kwargs: dict | None = None,
    reachability_kwargs: dict | None = None,
    sensitivity_kwargs: dict | None = None,
) -> tuple[Figure, tuple[ModeMatchingPlot, ReachabilityPlot, SensitivityPlot]]:
    """Plot the mode matching solution setup, reachability analysis, and sensitivity analysis in a single figure.

    Args:
        self: The mode matching solution instance.
        figsize: Figure size for the combined plot.
            If `None`, this defaults to :attr:`Config.PlotAll.figsize <corset.config.Config.PlotAll.figsize>`.
        width_ratios: Width ratios for the three subplots.
            If `None`, this defaults to :attr:`Config.PlotAll.width_ratios <corset.config.Config.PlotAll.width_ratios>`.
        tight_layout: Whether to use tight layout for the figure.
            If `None`, this defaults to :attr:`Config.PlotAll.tight_layout <corset.config.Config.PlotAll.tight_layout>`.
        setup_kwargs: Additional keyword arguments for the setup plot.
            If `None`, this defaults to :attr:`Config.PlotAll.setup_kwargs <corset.config.Config.PlotAll.setup_kwargs>`.
        reachability_kwargs: Additional keyword arguments for the reachability plot.
            If `None`, this defaults to :attr:`Config.PlotAll.reachability_kwargs <corset.config.Config.PlotAll.reachability_kwargs>`.
        sensitivity_kwargs: Additional keyword arguments for the sensitivity plot.
            If `None`, this defaults to :attr:`Config.PlotAll.sensitivity_kwargs <corset.config.Config.PlotAll.sensitivity_kwargs>`.

    Returns:
        A tuple containing the figure and an inner tuple with the three plot objects.
    """

    figsize = Config.get(figsize, Config.PlotAll.figsize)
    width_ratios = Config.get(width_ratios, Config.PlotAll.width_ratios)
    tight_layout = Config.get(tight_layout, Config.PlotAll.tight_layout)
    setup_kwargs = Config.get(setup_kwargs, Config.PlotAll.setup_kwargs)
    reachability_kwargs = Config.get(reachability_kwargs, Config.PlotAll.reachability_kwargs)
    sensitivity_kwargs = Config.get(sensitivity_kwargs, Config.PlotAll.sensitivity_kwargs)

    fig, (axl, axr, axc) = plt.subplots(
        1, 3, figsize=figsize, gridspec_kw={"width_ratios": width_ratios}, tight_layout=tight_layout
    )
    sol_plot = self.plot_setup(ax=axl, **setup_kwargs)
    reach_plot = self.plot_reachability(ax=axr, **reachability_kwargs)
    sens_plot = self.plot_sensitivity(ax=axc, **sensitivity_kwargs)
    return fig, (sol_plot, reach_plot, sens_plot)



from dataclasses import fields
from pathlib import Path
from typing import Any, Self

import numpy as np
import pandas as pd
import yaml


class YamlSerializableMixin:
    """Mixin class to add YAML serialization/deserialization to dataclasses and a few other common types."""

    class _Dumper(yaml.SafeDumper):
        pass

    class _Loader(yaml.SafeLoader):
        pass

    @classmethod
    def _register_misc_classes(cls) -> None:
        Dumper = cls._Dumper  # noqa: N806
        Loader = cls._Loader  # noqa: N806

        # python built-in classes
        @staticmethod
        def complex_representer(dumper: yaml.Dumper, data: complex) -> yaml.Node:
            return dumper.represent_sequence("!complex", [data.real, data.imag], flow_style=True)

        def complex_constructor(loader: yaml.Loader, node: yaml.Node) -> complex:
            return complex(*loader.construct_sequence(node, deep=True))  # pyright: ignore[reportArgumentType]

        Dumper.add_representer(complex, complex_representer)
        Loader.add_constructor("!complex", complex_constructor)  # pyright: ignore[reportArgumentType]

        def tuple_representer(dumper: yaml.Dumper, data: tuple) -> yaml.Node:
            return dumper.represent_sequence("!tuple", list(data))

        def tuple_constructor(loader: yaml.Loader, node: yaml.Node) -> tuple:
            return tuple(loader.construct_sequence(node, deep=True))  # pyright: ignore[reportArgumentType]

        Dumper.add_representer(tuple, tuple_representer)  # pyright: ignore[reportArgumentType]
        Loader.add_constructor("!tuple", tuple_constructor)  # pyright: ignore[reportArgumentType]

        # numpy classes
        def ndarray_representer(dumper: yaml.Dumper, data: np.ndarray) -> yaml.Node:
            return dumper.represent_sequence("!ndarray", data.tolist())

        def ndarray_constructor(loader: yaml.Loader, node: yaml.Node) -> np.ndarray:
            return np.array(loader.construct_sequence(node, deep=True))  # pyright: ignore[reportArgumentType]

        Dumper.add_representer(np.ndarray, ndarray_representer)  # pyright: ignore[reportArgumentType]
        Loader.add_constructor("!ndarray", ndarray_constructor)  # pyright: ignore[reportArgumentType]

        Dumper.add_representer(np.float32, lambda dumper, data: dumper.represent_float(float(data)))
        Dumper.add_representer(np.float64, lambda dumper, data: dumper.represent_float(float(data)))
        Dumper.add_representer(np.complex64, complex_representer)  # pyright: ignore[reportArgumentType]
        Dumper.add_representer(np.complex128, complex_representer)

    @staticmethod
    def __representer(dumper: yaml.Dumper, data: Any):
        # only store the dataclass fields
        attrs = {field.name: getattr(data, field.name) for field in fields(data)}
        return dumper.represent_mapping(f"!{type(data).__name__}", attrs)

    @classmethod
    def __constructor(cls, loader: yaml.Loader, node: yaml.Node):
        return cls(**loader.construct_mapping(node, deep=True))  # pyright: ignore[reportCallIssue, reportArgumentType]

    def __init_subclass__(cls) -> None:
        # access via base class not dynamic subclass
        YamlSerializableMixin._Dumper.add_representer(cls, cls.__representer)  # pyright: ignore[reportArgumentType]
        YamlSerializableMixin._Loader.add_constructor(f"!{cls.__name__}", cls.__constructor)

    def save_yaml(self: Any, filename: str | Path) -> None:
        """Save the object to a YAML file.

        Args:
            filename: Path to the YAML file.
        """
        from . import __version__

        yaml_data = {
            "meta": {"corset_version": __version__, "timestamp": pd.Timestamp.now().isoformat()},
            "data": self,
        }
        filename = Path(filename)
        filename.write_text(yaml.dump(yaml_data, Dumper=YamlSerializableMixin._Dumper, sort_keys=False))

    @classmethod
    def load_yaml(cls, filename: str | Path) -> Self:
        """Load an object from a YAML file.

        If this is called from a subclass the loaded object type must match that subclass.

        Args:
            filename: Path to the YAML file.

        Returns:
            The loaded object.

        Raises:
            ValueError: If the YAML file does not contain a 'data' field.
            TypeError: If the loaded object type does not match the class used to call this method.
        """

        filename = Path(filename)
        yaml_data = yaml.load(filename.read_text(), Loader=YamlSerializableMixin._Loader)  # noqa: S506
        if "data" not in yaml_data:
            raise ValueError("YAML file does not contain 'data' field.")
        data = yaml_data["data"]
        if cls is not YamlSerializableMixin and not isinstance(data, cls):
            raise TypeError(
                f"Attempting to load object of type '{type(data).__name__}' through different type '{cls.__name__}'."
            )
        return data


YamlSerializableMixin._register_misc_classes()

save_yaml = YamlSerializableMixin.save_yaml
"""Save any :class:`YamlSerializableMixin` instance or basic collection containing
them to a YAML file. This is an alias for :meth:`YamlSerializableMixin.save_yaml`.
"""
load_yaml = YamlSerializableMixin.load_yaml
"""Load any :class:`YamlSerializableMixin` instance or basic collection containing
them from a YAML file. This is an alias for :meth:`YamlSerializableMixin.load_yaml`."""
